import pickle
import datetime
import pytest
import uuid
from Aspidites._vendor.pyrsistent import (
    PRecord,
    field,
    InvariantException,
    ny,
    pset,
    PSet,
    CheckedPVector,
    PTypeError,
    pset_field,
    pvector_field,
    pmap_field,
    pmap,
    PMap,
    pvector,
    PVector,
    v,
    m,
)


class ARecord(PRecord):
    x = field(type=(int, float))
    y = field()


class Hierarchy(PRecord):
    point1 = field(ARecord)
    point2 = field(ARecord)
    points = pvector_field(ARecord)


class RecordContainingContainers(PRecord):
    map = pmap_field(str, str)
    vec = pvector_field(str)
    set = pset_field(str)


class UniqueThing(PRecord):
    id = field(type=uuid.UUID, factory=uuid.UUID)


class Something(object):
    pass


class Another(object):
    pass


def test_create_ignore_extra_true():
    h = Hierarchy.create(
        {
            "point1": {"x": 1, "y": "foo", "extra_field_0": "extra_data_0"},
            "point2": {"x": 1, "y": "foo", "extra_field_1": "extra_data_1"},
            "extra_field_2": "extra_data_2",
        },
        ignore_extra=True,
    )
    assert h


def test_create_ignore_extra_true_sequence_hierarchy():
    h = Hierarchy.create(
        {
            "point1": {"x": 1, "y": "foo", "extra_field_0": "extra_data_0"},
            "point2": {"x": 1, "y": "foo", "extra_field_1": "extra_data_1"},
            "points": [
                {"x": 1, "y": "foo", "extra_field_2": "extra_data_2"},
                {"x": 1, "y": "foo", "extra_field_3": "extra_data_3"},
            ],
            "extra_field____": "extra_data_2",
        },
        ignore_extra=True,
    )
    assert h


def test_ignore_extra_for_pvector_field():
    class HierarchyA(PRecord):
        points = pvector_field(ARecord, optional=False)

    class HierarchyB(PRecord):
        points = pvector_field(ARecord, optional=True)

    point_object = {"x": 1, "y": "foo", "extra_field": 69}

    h = HierarchyA.create({"points": [point_object]}, ignore_extra=True)
    assert h
    h = HierarchyB.create({"points": [point_object]}, ignore_extra=True)
    assert h


def test_create():
    r = ARecord(x=1, y="foo")
    assert r.x == 1
    assert r.y == "foo"
    assert isinstance(r, ARecord)


def test_create_ignore_extra():
    r = ARecord.create({"x": 1, "y": "foo", "z": None}, ignore_extra=True)
    assert r.x == 1
    assert r.y == "foo"
    assert isinstance(r, ARecord)


def test_create_ignore_extra_false():
    with pytest.raises(AttributeError):
        _ = ARecord.create({"x": 1, "y": "foo", "z": None})


def test_correct_assignment():
    r = ARecord(x=1, y="foo")
    r2 = r.set("x", 2.0)
    r3 = r2.set("y", "bar")

    assert r2 == {"x": 2.0, "y": "foo"}
    assert r3 == {"x": 2.0, "y": "bar"}
    assert isinstance(r3, ARecord)


def test_direct_assignment_not_possible():
    with pytest.raises(AttributeError):
        ARecord().x = 1


def test_cannot_assign_undeclared_fields():
    with pytest.raises(AttributeError):
        ARecord().set("z", 5)


def test_cannot_assign_wrong_type_to_fields():
    try:
        ARecord().set("x", "foo")
        assert False
    except PTypeError as e:
        assert e.source_class == ARecord
        assert e.field == "x"
        assert e.expected_types == set([int, float])
        assert e.actual_type is type("foo")


def test_cannot_construct_with_undeclared_fields():
    with pytest.raises(AttributeError):
        ARecord(z=5)


def test_cannot_construct_with_fields_of_wrong_type():
    with pytest.raises(TypeError):
        ARecord(x="foo")


def test_support_record_inheritance():
    class BRecord(ARecord):
        z = field()

    r = BRecord(x=1, y="foo", z="bar")

    assert isinstance(r, BRecord)
    assert isinstance(r, ARecord)
    assert r == {"x": 1, "y": "foo", "z": "bar"}


def test_single_type_spec():
    class A(PRecord):
        x = field(type=int)

    r = A(x=1)
    assert r.x == 1

    with pytest.raises(TypeError):
        r.set("x", "foo")


def test_remove():
    r = ARecord(x=1, y="foo")
    r2 = r.remove("y")

    assert isinstance(r2, ARecord)
    assert r2 == {"x": 1}


def test_remove_non_existing_member():
    r = ARecord(x=1, y="foo")

    with pytest.raises(KeyError):
        r.remove("z")


def test_field_invariant_must_hold():
    class BRecord(PRecord):
        x = field(invariant=lambda x: (x > 1, "x too small"))
        y = field(mandatory=True)

    try:
        BRecord(x=1)
        assert False
    except InvariantException as e:
        assert e.invariant_errors == ("x too small",)
        assert e.missing_fields == ("BRecord.y",)


def test_global_invariant_must_hold():
    class BRecord(PRecord):
        __invariant__ = lambda r: (r.x <= r.y, "y smaller than x")
        x = field()
        y = field()

    BRecord(x=1, y=2)

    try:
        BRecord(x=2, y=1)
        assert False
    except InvariantException as e:
        assert e.invariant_errors == ("y smaller than x",)
        assert e.missing_fields == ()


def test_set_multiple_fields():
    a = ARecord(x=1, y="foo")
    b = a.set(x=2, y="bar")

    assert b == {"x": 2, "y": "bar"}


def test_initial_value():
    class BRecord(PRecord):
        x = field(initial=1)
        y = field(initial=2)

    a = BRecord()
    assert a.x == 1
    assert a.y == 2


def test_enum_field():
    try:
        from enum import Enum
    except ImportError:
        return  # Enum not supported in this environment

    class ExampleEnum(Enum):
        x = 1
        y = 2

    class RecordContainingEnum(PRecord):
        enum_field = field(type=ExampleEnum)

    r = RecordContainingEnum(enum_field=ExampleEnum.x)
    assert r.enum_field == ExampleEnum.x


def test_type_specification_must_be_a_type():
    with pytest.raises(TypeError):

        class BRecord(PRecord):
            x = field(type=1)


def test_initial_must_be_of_correct_type():
    with pytest.raises(TypeError):

        class BRecord(PRecord):
            x = field(type=int, initial="foo")


def test_invariant_must_be_callable():
    with pytest.raises(TypeError):

        class BRecord(PRecord):
            x = field(invariant="foo")  # type: ignore


def test_global_invariants_are_inherited():
    class BRecord(PRecord):
        __invariant__ = lambda r: (r.x % r.y == 0, "modulo")
        x = field()
        y = field()

    class CRecord(BRecord):
        __invariant__ = lambda r: (r.x > r.y, "size")

    try:
        CRecord(x=5, y=3)
        assert False
    except InvariantException as e:
        assert e.invariant_errors == ("modulo",)


def test_global_invariants_must_be_callable():
    with pytest.raises(TypeError):

        class CRecord(PRecord):
            __invariant__ = 1


def test_repr():
    r = ARecord(x=1, y=2)
    assert repr(r) == "ARecord(x=1, y=2)" or repr(r) == "ARecord(y=2, x=1)"


def test_factory():
    class BRecord(PRecord):
        x = field(type=int, factory=int)

    assert BRecord(x=2.5) == {"x": 2}


def test_factory_must_be_callable():
    with pytest.raises(TypeError):

        class BRecord(PRecord):
            x = field(type=int, factory=1)  # type: ignore


def test_nested_record_construction():
    class BRecord(PRecord):
        x = field(int, factory=int)

    class CRecord(PRecord):
        a = field()
        b = field(type=BRecord)

    r = CRecord.create({"a": "foo", "b": {"x": "5"}})
    assert isinstance(r, CRecord)
    assert isinstance(r.b, BRecord)
    assert r == {"a": "foo", "b": {"x": 5}}


def test_pickling():
    x = ARecord(x=2.0, y="bar")
    y = pickle.loads(pickle.dumps(x, -1))

    assert x == y
    assert isinstance(y, ARecord)


def test_supports_pickling_with_typed_container_fields():
    obj = RecordContainingContainers(
        map={"foo": "bar"}, set=["hello", "there"], vec=["a", "b"]
    )
    obj2 = pickle.loads(pickle.dumps(obj))
    assert obj == obj2


def test_all_invariant_errors_reported():
    class BRecord(PRecord):
        x = field(factory=int, invariant=lambda x: (x >= 0, "x negative"))
        y = field(mandatory=True)

    class CRecord(PRecord):
        a = field(invariant=lambda x: (x != 0, "a zero"))
        b = field(type=BRecord)

    try:
        CRecord.create({"a": 0, "b": {"x": -5}})
        assert False
    except InvariantException as e:
        assert set(e.invariant_errors) == set(["x negative", "a zero"])
        assert e.missing_fields == ("BRecord.y",)


def test_precord_factory_method_is_idempotent():
    class BRecord(PRecord):
        x = field()
        y = field()

    r = BRecord(x=1, y=2)
    assert BRecord.create(r) is r


def test_serialize():
    class BRecord(PRecord):
        d = field(
            type=datetime.date,
            factory=lambda d: datetime.datetime.strptime(d, "%d%m%Y").date(),
            serializer=lambda format, d: d.strftime("%Y-%m-%d")
            if format == "ISO"
            else d.strftime("%d%m%Y"),
        )

    assert BRecord(d="14012015").serialize("ISO") == {"d": "2015-01-14"}
    assert BRecord(d="14012015").serialize("other") == {"d": "14012015"}


def test_nested_serialize():
    class BRecord(PRecord):
        d = field(serializer=lambda format, d: format)

    class CRecord(PRecord):
        b = field()

    serialized = CRecord(b=BRecord(d="foo")).serialize("bar")

    assert serialized == {"b": {"d": "bar"}}
    assert isinstance(serialized, dict)


def test_serializer_must_be_callable():
    with pytest.raises(TypeError):

        class CRecord(PRecord):
            x = field(serializer=1)  # type: ignore


def test_transform_without_update_returns_same_precord():
    r = ARecord(x=2.0, y="bar")
    assert r.transform([ny], lambda x: x) is r


class Application(PRecord):
    name = field(type=str)
    image = field(type=str)


class ApplicationVector(CheckedPVector):
    __type__ = Application


class Node(PRecord):
    applications = field(type=ApplicationVector)


def test_nested_create_serialize():
    node = Node(
        applications=[
            Application(name="myapp", image="myimage"),
            Application(name="b", image="c"),
        ]
    )

    node2 = Node.create(
        {
            "applications": [
                {"name": "myapp", "image": "myimage"},
                {"name": "b", "image": "c"},
            ]
        }
    )

    assert node == node2

    serialized = node.serialize()
    restored = Node.create(serialized)

    assert restored == node


def test_pset_field_initial_value():
    """
    ``pset_field`` results in initial value that is empty.
    """

    class Record(PRecord):
        value = pset_field(int)

    assert Record() == Record(value=[])


def test_pset_field_custom_initial():
    """
    A custom initial value can be passed in.
    """

    class Record(PRecord):
        value = pset_field(int, initial=(1, 2))

    assert Record() == Record(value=[1, 2])


def test_pset_field_factory():
    """
    ``pset_field`` has a factory that creates a ``PSet``.
    """

    class Record(PRecord):
        value = pset_field(int)

    record = Record(value=[1, 2])
    assert isinstance(record.value, PSet)


def test_pset_field_checked_set():
    """
    ``pset_field`` results in a set that enforces its type.
    """

    class Record(PRecord):
        value = pset_field(int)

    record = Record(value=[1, 2])
    with pytest.raises(TypeError):
        record.value.add("hello")  # type: ignore


def test_pset_field_checked_vector_multiple_types():
    """
    ``pset_field`` results in a vector that enforces its types.
    """

    class Record(PRecord):
        value = pset_field((int, str))

    record = Record(value=[1, 2, "hello"])
    with pytest.raises(TypeError):
        record.value.add(object())


def test_pset_field_type():
    """
    ``pset_field`` enforces its type.
    """

    class Record(PRecord):
        value = pset_field(int)

    record = Record()
    with pytest.raises(TypeError):
        record.set("value", None)


def test_pset_field_mandatory():
    """
    ``pset_field`` is a mandatory field.
    """

    class Record(PRecord):
        value = pset_field(int)

    record = Record(value=[1])
    with pytest.raises(InvariantException):
        record.remove("value")


def test_pset_field_default_non_optional():
    """
    By default ``pset_field`` is non-optional, i.e. does not allow
    ``None``.
    """

    class Record(PRecord):
        value = pset_field(int)

    with pytest.raises(TypeError):
        Record(value=None)


def test_pset_field_explicit_non_optional():
    """
    If ``optional`` argument is ``False`` then ``pset_field`` is
    non-optional, i.e. does not allow ``None``.
    """

    class Record(PRecord):
        value = pset_field(int, optional=False)

    with pytest.raises(TypeError):
        Record(value=None)


def test_pset_field_optional():
    """
    If ``optional`` argument is true, ``None`` is acceptable alternative
    to a set.
    """

    class Record(PRecord):
        value = pset_field(int, optional=True)

    assert (Record(value=[1, 2]).value, Record(value=None).value) == (
        pset([1, 2]),
        None,
    )


def test_pset_field_name():
    """
    The created set class name is based on the type of items in the set.
    """

    class Record(PRecord):
        value = pset_field(Something)
        value2 = pset_field(int)

    assert (Record().value.__class__.__name__, Record().value2.__class__.__name__) == (
        "SomethingPSet",
        "IntPSet",
    )


def test_pset_multiple_types_field_name():
    """
    The created set class name is based on the multiple given types of
    items in the set.
    """

    class Record(PRecord):
        value = pset_field((Something, int))

    assert Record().value.__class__.__name__ == "SomethingIntPSet"


def test_pset_field_name_string_type():
    """
    The created set class name is based on the type of items specified by name
    """

    class Record(PRecord):
        value = pset_field("record_test.Something")

    assert Record().value.__class__.__name__ == "SomethingPSet"


def test_pset_multiple_string_types_field_name():
    """
    The created set class name is based on the multiple given types of
    items in the set specified by name
    """

    class Record(PRecord):
        value = pset_field(("record_test.Something", "record_test.Another"))

    assert Record().value.__class__.__name__ == "SomethingAnotherPSet"


def test_pvector_field_initial_value():
    """
    ``pvector_field`` results in initial value that is empty.
    """

    class Record(PRecord):
        value = pvector_field(int)

    assert Record() == Record(value=[])


def test_pvector_field_custom_initial():
    """
    A custom initial value can be passed in.
    """

    class Record(PRecord):
        value = pvector_field(int, initial=(1, 2))

    assert Record() == Record(value=[1, 2])


def test_pvector_field_factory():
    """
    ``pvector_field`` has a factory that creates a ``PVector``.
    """

    class Record(PRecord):
        value = pvector_field(int)

    record = Record(value=[1, 2])
    assert isinstance(record.value, PVector)


def test_pvector_field_checked_vector():
    """
    ``pvector_field`` results in a vector that enforces its type.
    """

    class Record(PRecord):
        value = pvector_field(int)

    record = Record(value=[1, 2])
    with pytest.raises(TypeError):
        record.value.append("hello")  # type: ignore


def test_pvector_field_checked_vector_multiple_types():
    """
    ``pvector_field`` results in a vector that enforces its types.
    """

    class Record(PRecord):
        value = pvector_field((int, str))

    record = Record(value=[1, 2, "hello"])
    with pytest.raises(TypeError):
        record.value.append(object())


def test_pvector_field_type():
    """
    ``pvector_field`` enforces its type.
    """

    class Record(PRecord):
        value = pvector_field(int)

    record = Record()
    with pytest.raises(TypeError):
        record.set("value", None)


def test_pvector_field_mandatory():
    """
    ``pvector_field`` is a mandatory field.
    """

    class Record(PRecord):
        value = pvector_field(int)

    record = Record(value=[1])
    with pytest.raises(InvariantException):
        record.remove("value")


def test_pvector_field_default_non_optional():
    """
    By default ``pvector_field`` is non-optional, i.e. does not allow
    ``None``.
    """

    class Record(PRecord):
        value = pvector_field(int)

    with pytest.raises(TypeError):
        Record(value=None)


def test_pvector_field_explicit_non_optional():
    """
    If ``optional`` argument is ``False`` then ``pvector_field`` is
    non-optional, i.e. does not allow ``None``.
    """

    class Record(PRecord):
        value = pvector_field(int, optional=False)

    with pytest.raises(TypeError):
        Record(value=None)


def test_pvector_field_optional():
    """
    If ``optional`` argument is true, ``None`` is acceptable alternative
    to a sequence.
    """

    class Record(PRecord):
        value = pvector_field(int, optional=True)

    assert (Record(value=[1, 2]).value, Record(value=None).value) == (
        pvector([1, 2]),
        None,
    )


def test_pvector_field_name():
    """
    The created set class name is based on the type of items in the set.
    """

    class Record(PRecord):
        value = pvector_field(Something)
        value2 = pvector_field(int)

    assert (Record().value.__class__.__name__, Record().value2.__class__.__name__) == (
        "SomethingPVector",
        "IntPVector",
    )


def test_pvector_multiple_types_field_name():
    """
    The created vector class name is based on the multiple given types of
    items in the vector.
    """

    class Record(PRecord):
        value = pvector_field((Something, int))

    assert Record().value.__class__.__name__ == "SomethingIntPVector"


def test_pvector_field_name_string_type():
    """
    The created set class name is based on the type of items in the set
    specified by name.
    """

    class Record(PRecord):
        value = pvector_field("record_test.Something")

    assert Record().value.__class__.__name__ == "SomethingPVector"


def test_pvector_multiple_string_types_field_name():
    """
    The created vector class name is based on the multiple given types of
    items in the vector.
    """

    class Record(PRecord):
        value = pvector_field(("record_test.Something", "record_test.Another"))

    assert Record().value.__class__.__name__ == "SomethingAnotherPVector"


def test_pvector_field_create_from_nested_serialized_data():
    class Foo(PRecord):
        foo = field(type=str)

    class Bar(PRecord):
        bar = pvector_field(Foo)

    data = Bar(bar=v(Foo(foo="foo")))
    Bar.create(data.serialize()) == data


def test_pmap_field_initial_value():
    """
    ``pmap_field`` results in initial value that is empty.
    """

    class Record(PRecord):
        value = pmap_field(int, int)

    assert Record() == Record(value={})


def test_pmap_field_factory():
    """
    ``pmap_field`` has a factory that creates a ``PMap``.
    """

    class Record(PRecord):
        value = pmap_field(int, int)

    record = Record(value={1: 1234})
    assert isinstance(record.value, PMap)


def test_pmap_field_checked_map_key():
    """
    ``pmap_field`` results in a map that enforces its key type.
    """

    class Record(PRecord):
        value = pmap_field(int, type(None))

    record = Record(value={1: None})
    with pytest.raises(TypeError):
        record.value.set("hello", None)  # type: ignore


def test_pmap_field_checked_map_value():
    """
    ``pmap_field`` results in a map that enforces its value type.
    """

    class Record(PRecord):
        value = pmap_field(int, type(None))

    record = Record(value={1: None})
    with pytest.raises(TypeError):
        record.value.set(2, 4)  # type: ignore


def test_pmap_field_checked_map_key_multiple_types():
    """
    ``pmap_field`` results in a map that enforces its key types.
    """

    class Record(PRecord):
        value = pmap_field((int, str), type(None))

    record = Record(value={1: None, "hello": None})
    with pytest.raises(TypeError):
        record.value.set(object(), None)


def test_pmap_field_checked_map_value_multiple_types():
    """
    ``pmap_field`` results in a map that enforces its value types.
    """

    class Record(PRecord):
        value = pmap_field(int, (str, type(None)))

    record = Record(value={1: None, 3: "hello"})
    with pytest.raises(TypeError):
        record.value.set(2, 4)


def test_pmap_field_mandatory():
    """
    ``pmap_field`` is a mandatory field.
    """

    class Record(PRecord):
        value = pmap_field(int, int)

    record = Record()
    with pytest.raises(InvariantException):
        record.remove("value")


def test_pmap_field_default_non_optional():
    """
    By default ``pmap_field`` is non-optional, i.e. does not allow
    ``None``.
    """

    class Record(PRecord):
        value = pmap_field(int, int)

    # Ought to be TypeError, but pyrsistent doesn't quite allow that:
    with pytest.raises(AttributeError):
        Record(value=None)


def test_pmap_field_explicit_non_optional():
    """
    If ``optional`` argument is ``False`` then ``pmap_field`` is
    non-optional, i.e. does not allow ``None``.
    """

    class Record(PRecord):
        value = pmap_field(int, int, optional=False)

    # Ought to be TypeError, but pyrsistent doesn't quite allow that:
    with pytest.raises(AttributeError):
        Record(value=None)


def test_pmap_field_optional():
    """
    If ``optional`` argument is true, ``None`` is acceptable alternative
    to a set.
    """

    class Record(PRecord):
        value = pmap_field(int, int, optional=True)

    assert (Record(value={1: 2}).value, Record(value=None).value) == (
        pmap({1: 2}),
        None,
    )


def test_pmap_field_name():
    """
    The created map class name is based on the types of items in the map.
    """

    class Record(PRecord):
        value = pmap_field(Something, Another)
        value2 = pmap_field(int, float)

    assert (Record().value.__class__.__name__, Record().value2.__class__.__name__) == (
        "SomethingToAnotherPMap",
        "IntToFloatPMap",
    )


def test_pmap_field_name_multiple_types():
    """
    The created map class name is based on the types of items in the map,
    including when there are multiple supported types.
    """

    class Record(PRecord):
        value = pmap_field((Something, Another), int)
        value2 = pmap_field(str, (int, float))

    assert (Record().value.__class__.__name__, Record().value2.__class__.__name__) == (
        "SomethingAnotherToIntPMap",
        "StrToIntFloatPMap",
    )


def test_pmap_field_name_string_type():
    """
    The created map class name is based on the types of items in the map
    specified by name.
    """

    class Record(PRecord):
        value = pmap_field("record_test.Something", "record_test.Another")

    assert Record().value.__class__.__name__ == "SomethingToAnotherPMap"


def test_pmap_field_name_multiple_string_types():
    """
    The created map class name is based on the types of items in the map,
    including when there are multiple supported types.
    """

    class Record(PRecord):
        value = pmap_field(("record_test.Something", "record_test.Another"), int)
        value2 = pmap_field(str, ("record_test.Something", "record_test.Another"))

    assert (Record().value.__class__.__name__, Record().value2.__class__.__name__) == (
        "SomethingAnotherToIntPMap",
        "StrToSomethingAnotherPMap",
    )


def test_pmap_field_invariant():
    """
    The ``invariant`` parameter is passed through to ``field``.
    """

    class Record(PRecord):
        value = pmap_field(
            int,
            int,
            invariant=(lambda pmap: (len(pmap) == 1, "Exactly one item required.")),
        )

    with pytest.raises(InvariantException):
        Record(value={})
    with pytest.raises(InvariantException):
        Record(value={1: 2, 3: 4})
    assert Record(value={1: 2}).value == {1: 2}


def test_pmap_field_create_from_nested_serialized_data():
    class Foo(PRecord):
        foo = field(type=str)

    class Bar(PRecord):
        bar = pmap_field(str, Foo)

    data = Bar(bar=m(foo_key=Foo(foo="foo")))
    Bar.create(data.serialize()) == data


def test_supports_weakref():
    import weakref

    weakref.ref(ARecord(x=1, y=2))


def test_supports_lazy_initial_value_for_field():
    class MyRecord(PRecord):
        a = field(int, initial=lambda: 2)

    assert MyRecord() == MyRecord(a=2)


def test_pickle_with_one_way_factory():
    """
    A field factory isn't called when restoring from pickle.
    """
    thing = UniqueThing(id="25544626-86da-4bce-b6b6-9186c0804d64")
    assert thing == pickle.loads(pickle.dumps(thing))



# select_parser.py
# Copyright 2010,2019 Paul McGuire
#
# a simple SELECT statement parser, taken from SQLite's SELECT statement
# definition at https://www.sqlite.org/lang_select.html
#
from pyparsing import *

ParserElement.enablePackrat()

LPAR, RPAR, COMMA = map(Suppress, "(),")
DOT, STAR = map(Literal, ".*")
select_stmt = Forward().setName("select statement")

# keywords
(
    UNION,
    ALL,
    AND,
    INTERSECT,
    EXCEPT,
    COLLATE,
    ASC,
    DESC,
    ON,
    USING,
    NATURAL,
    INNER,
    CROSS,
    LEFT,
    OUTER,
    JOIN,
    AS,
    INDEXED,
    NOT,
    SELECT,
    DISTINCT,
    FROM,
    WHERE,
    GROUP,
    BY,
    HAVING,
    ORDER,
    BY,
    LIMIT,
    OFFSET,
    OR,
) = map(
    CaselessKeyword,
    """UNION, ALL, AND, INTERSECT,
 EXCEPT, COLLATE, ASC, DESC, ON, USING, NATURAL, INNER, CROSS, LEFT, OUTER, JOIN, AS, INDEXED, NOT, SELECT,
 DISTINCT, FROM, WHERE, GROUP, BY, HAVING, ORDER, BY, LIMIT, OFFSET, OR""".replace(
        ",", ""
    ).split(),
)
(
    CAST,
    ISNULL,
    NOTNULL,
    NULL,
    IS,
    BETWEEN,
    ELSE,
    END,
    CASE,
    WHEN,
    THEN,
    EXISTS,
    IN,
    LIKE,
    GLOB,
    REGEXP,
    MATCH,
    ESCAPE,
    CURRENT_TIME,
    CURRENT_DATE,
    CURRENT_TIMESTAMP,
) = map(
    CaselessKeyword,
    """CAST, ISNULL, NOTNULL, NULL, IS, BETWEEN, ELSE, END, CASE, WHEN, THEN, EXISTS, IN, LIKE, GLOB, 
 REGEXP, MATCH, ESCAPE, CURRENT_TIME, CURRENT_DATE, CURRENT_TIMESTAMP""".replace(
        ",", ""
    ).split(),
)
keyword = MatchFirst(
    (
        UNION,
        ALL,
        INTERSECT,
        EXCEPT,
        COLLATE,
        ASC,
        DESC,
        ON,
        USING,
        NATURAL,
        INNER,
        CROSS,
        LEFT,
        OUTER,
        JOIN,
        AS,
        INDEXED,
        NOT,
        SELECT,
        DISTINCT,
        FROM,
        WHERE,
        GROUP,
        BY,
        HAVING,
        ORDER,
        BY,
        LIMIT,
        OFFSET,
        CAST,
        ISNULL,
        NOTNULL,
        NULL,
        IS,
        BETWEEN,
        ELSE,
        END,
        CASE,
        WHEN,
        THEN,
        EXISTS,
        COLLATE,
        IN,
        LIKE,
        GLOB,
        REGEXP,
        MATCH,
        ESCAPE,
        CURRENT_TIME,
        CURRENT_DATE,
        CURRENT_TIMESTAMP,
    )
)

identifier = ~keyword + Word(alphas, alphanums + "_")
collation_name = identifier.copy()
column_name = identifier.copy()
column_alias = identifier.copy()
table_name = identifier.copy()
table_alias = identifier.copy()
index_name = identifier.copy()
function_name = identifier.copy()
parameter_name = identifier.copy()
database_name = identifier.copy()

# expression
expr = Forward().setName("expression")

integer = Regex(r"[+-]?\d+")
numeric_literal = Regex(r"\d+(\.\d*)?([eE][+-]?\d+)?")
string_literal = QuotedString("'")
blob_literal = Regex(r"[xX]'[0-9A-Fa-f]+'")
literal_value = (
    numeric_literal
    | string_literal
    | blob_literal
    | NULL
    | CURRENT_TIME
    | CURRENT_DATE
    | CURRENT_TIMESTAMP
)
bind_parameter = Word("?", nums) | Combine(oneOf(": @ $") + parameter_name)
type_name = oneOf("TEXT REAL INTEGER BLOB NULL")

expr_term = (
    CAST + LPAR + expr + AS + type_name + RPAR
    | EXISTS + LPAR + select_stmt + RPAR
    | function_name.setName("function_name")
    + LPAR
    + Optional(STAR | delimitedList(expr))
    + RPAR
    | literal_value
    | bind_parameter
    | Group(
        identifier("col_db") + DOT + identifier("col_tab") + DOT + identifier("col")
    )
    | Group(identifier("col_tab") + DOT + identifier("col"))
    | Group(identifier("col"))
)

UNARY, BINARY, TERNARY = 1, 2, 3
expr << infixNotation(
    expr_term,
    [
        (oneOf("- + ~") | NOT, UNARY, opAssoc.RIGHT),
        (ISNULL | NOTNULL | NOT + NULL, UNARY, opAssoc.LEFT),
        ("||", BINARY, opAssoc.LEFT),
        (oneOf("* / %"), BINARY, opAssoc.LEFT),
        (oneOf("+ -"), BINARY, opAssoc.LEFT),
        (oneOf("<< >> & |"), BINARY, opAssoc.LEFT),
        (oneOf("< <= > >="), BINARY, opAssoc.LEFT),
        (
            oneOf("= == != <>") | IS | IN | LIKE | GLOB | MATCH | REGEXP,
            BINARY,
            opAssoc.LEFT,
        ),
        ((BETWEEN, AND), TERNARY, opAssoc.LEFT),
        (
            IN + LPAR + Group(select_stmt | delimitedList(expr)) + RPAR,
            UNARY,
            opAssoc.LEFT,
        ),
        (AND, BINARY, opAssoc.LEFT),
        (OR, BINARY, opAssoc.LEFT),
    ],
)

compound_operator = UNION + Optional(ALL) | INTERSECT | EXCEPT

ordering_term = Group(
    expr("order_key")
    + Optional(COLLATE + collation_name("collate"))
    + Optional(ASC | DESC)("direction")
)

join_constraint = Group(
    Optional(ON + expr | USING + LPAR + Group(delimitedList(column_name)) + RPAR)
)

join_op = COMMA | Group(
    Optional(NATURAL) + Optional(INNER | CROSS | LEFT + OUTER | LEFT | OUTER) + JOIN
)

join_source = Forward()
single_source = (
    Group(database_name("database") + DOT + table_name("table*") | table_name("table*"))
    + Optional(Optional(AS) + table_alias("table_alias*"))
    + Optional(INDEXED + BY + index_name("name") | NOT + INDEXED)("index")
    | (LPAR + select_stmt + RPAR + Optional(Optional(AS) + table_alias))
    | (LPAR + join_source + RPAR)
)

join_source <<= (
    Group(single_source + OneOrMore(join_op + single_source + join_constraint))
    | single_source
)

# result_column = "*" | table_name + "." + "*" | Group(expr + Optional(Optional(AS) + column_alias))
result_column = Group(
    STAR("col")
    | table_name("col_table") + DOT + STAR("col")
    | expr("col") + Optional(Optional(AS) + column_alias("alias"))
)

select_core = (
    SELECT
    + Optional(DISTINCT | ALL)
    + Group(delimitedList(result_column))("columns")
    + Optional(FROM + join_source("from*"))
    + Optional(WHERE + expr("where_expr"))
    + Optional(
        GROUP
        + BY
        + Group(delimitedList(ordering_term))("group_by_terms")
        + Optional(HAVING + expr("having_expr"))
    )
)

select_stmt << (
    select_core
    + ZeroOrMore(compound_operator + select_core)
    + Optional(ORDER + BY + Group(delimitedList(ordering_term))("order_by_terms"))
    + Optional(
        LIMIT
        + (Group(expr + OFFSET + expr) | Group(expr + COMMA + expr) | expr)("limit")
    )
)

tests = """\
    select * from xyzzy where z > 100
    select * from xyzzy where z > 100 order by zz
    select * from xyzzy
    select z.* from xyzzy
    select a, b from test_table where 1=1 and b='yes'
    select a, b from test_table where 1=1 and b in (select bb from foo)
    select z.a, b from test_table where 1=1 and b in (select bb from foo)
    select z.a, b from test_table where 1=1 and b in (select bb from foo) order by b,c desc,d
    select z.a, b from test_table left join test2_table where 1=1 and b in (select bb from foo)
    select a, db.table.b as BBB from db.table where 1=1 and BBB='yes'
    select a, db.table.b as BBB from test_table,db.table where 1=1 and BBB='yes'
    select a, db.table.b as BBB from test_table,db.table where 1=1 and BBB='yes' limit 50
    select a, b from test_table where (1=1 or 2=3) and b='yes' group by zx having b=2 order by 1
    SELECT emp.ename as e FROM scott.employee as emp
    SELECT ename as e, fname as f FROM scott.employee as emp
    SELECT emp.eid, fname,lname FROM scott.employee as emp
    SELECT ename, lname, emp.eid FROM scott.employee as emp
    select emp.salary * (1.0 + emp.bonus) as salary_plus_bonus from scott.employee as emp
"""
select_stmt.runTests(tests)



import unittest

from Aspidites._vendor.contracts import new_contract, check, Contract, contract
from Aspidites._vendor.contracts.library.extensions import identifier_expression

from .utils import check_contracts_fail, check_contracts_ok
from Aspidites._vendor.contracts.main_actual import Storage
from Aspidites._vendor.contracts.syntax import ParsingTmp

# The different patterns


def ok1(x):
    pass


def ok2(x):  # @UnusedVariable
    return True


def fail1(x):
    raise ValueError("message")


def fail2(x):  # @UnusedVariable
    return False


def invalid_callable1(x):  # @UnusedVariable
    return "ciao"


# generates a new name:
def cname():
    TestNewContract.counter += 1
    return "GeneratedContract%d" % TestNewContract.counter


class TestNewContract(unittest.TestCase):
    counter = 0

    # def setup(self):
    #    Storage.string2contract = {}

    def test_inverted_args(self):
        self.assertRaises(ValueError, new_contract, ok1, "list")

    def test_wrong_args(self):
        self.assertRaises(ValueError, new_contract, "my13", 2)

    def test_invalid_callable(self):
        self.assertRaises(ValueError, new_contract, "new", lambda: None)

    def test_parsing_error(self):
        self.assertRaises(ValueError, new_contract, "new", ">>")

    def test_parsing_error2(self):
        # parsing error (unknown spec)
        self.assertRaises(ValueError, new_contract, "new", "unknown")

    def test_invalid_names(self):
        # invalid names:
        alphabet = "A B C D E F G H I J K L M N O P Q R S T U W V X Y Z"
        for x in alphabet.split():
            self.assertRaises(ValueError, new_contract, x, "list")
            self.assertRaises(ValueError, new_contract, x.lower(), "list")
        self.assertRaises(ValueError, new_contract, "list", "list[N]")
        self.assertRaises(ValueError, new_contract, "2acdca", "list[N]")
        self.assertRaises(ValueError, new_contract, "_", "list[N]")

    def test_valid(self):
        c = new_contract("my_list", "list[2]")
        assert isinstance(c, Contract)
        check("tuple(my_list, my_list)", ([1, 2], [1, 2]))
        check_contracts_fail("tuple(my_list, my_list)", ([1, 2], [1, 2, 3]))

    def test_separate_context(self):
        new_contract("my_list2", "list[N]")
        check("tuple(my_list2, my_list2)", ([1, 2], [1, 2]))
        check("tuple(my_list2, my_list2)", ([1, 2], [1, 2, 3]))

    def test_renaming(self):
        self.assertNotEqual(ok1, ok2)
        new_contract("my7", ok1)
        self.assertRaises(ValueError, new_contract, "my7", ok2)

    def test_allow_renaming_if_equal1(self):
        new_contract("my8", ok1)
        new_contract("my8", ok1)

    def test_allow_renaming_if_equal2(self):
        new_contract("my8b", "list[3]")
        new_contract("my8b", "list[3]")

    def test_callable1(self):
        c = cname()
        new_contract(c, ok2)
        check("list(%s)" % c, [0])

    def test_callable2(self):
        c = cname()
        new_contract(c, ok2)
        check("list(%s)" % c, [0])

    def test_callable3(self):
        c = cname()
        new_contract(c, fail1)
        check_contracts_fail("list(%s)" % c, [0])

    def test_callable4(self):
        c = cname()
        new_contract(c, fail2)
        check_contracts_fail("list(%s)" % c, [0])

    def test_invalid_callable2(self):
        c = cname()
        new_contract(c, invalid_callable1)
        self.assertRaises(ValueError, check, "list(%s)" % c, [0])

    def test_other_pass(self):
        class Ex1(Exception):
            pass

        def invalid(x):
            raise Ex1()

        c = cname()
        new_contract(c, invalid)
        self.assertRaises(Ex1, check, "list(%s)" % c, [0])

    def test_callable(self):
        class MyTest_ok(object):
            def __call__(self, x):  # @UnusedVariable
                return True

        o = MyTest_ok()
        assert o("value") == True
        new_contract(cname(), o)

    def test_callable_old_style(self):
        class MyTest_ok:
            def __call__(self, x):  # @UnusedVariable
                return True

        o = MyTest_ok()
        assert o("value") == True
        new_contract(cname(), o)

    def test_class_as_contract1(self):
        # This should be interpreted as a type
        # init(x,y) so it is not mistaken for a valid callable
        class NewStyleClass(object):
            def __init__(self, x, y):  # @UnusedVariable
                pass

        new_contract(cname(), NewStyleClass)

    def test_class_as_contract2(self):
        # old sytle class
        class OldStyleClass:
            def __init__(self, x, y):  # @UnusedVariable
                pass

        new_contract(cname(), OldStyleClass)

    def test_class_as_contract3(self):
        class NewStyleClass(object):
            def __init__(self, x, y):  # @UnusedVariable
                pass

        @contract(x=NewStyleClass)
        def f(x):
            pass

    def test_class_as_contract4(self):
        class OldStyleClass:
            def __init__(self, x, y):  # @UnusedVariable
                pass

        @contract(x=OldStyleClass)
        def f(x):
            pass

    def test_callable_5(self):
        class MyTest_ok(object):
            def f(self, x):  # @UnusedVariable
                return True

        o = MyTest_ok()
        assert o.f("value") == True
        new_contract(cname(), o.f)

    # removed after checking in class methods
    #    def test_callable_invalid(self):
    #        class MyTest_fail(object):
    #            def __call__(self, x, y):  # @UnusedVariable
    #                return True
    #
    #        self.assertRaises(ValueError, new_contract, cname(), MyTest_fail())

    def test_lambda_2(self):
        new_contract(cname(), lambda x: True)  # @UnusedVariable
        new_contract(cname(), lambda x: None)  # @UnusedVariable

    def test_lambda_invalid(self):
        f = lambda: True  # @UnusedVariable
        self.assertRaises(ValueError, new_contract, cname(), f)

    def test_idioms(self):
        Storage.string2contract = {}  # TODO remove this hack
        color = new_contract("color", "list[3](number,>=0,<=1)")
        # Make sure we got it right
        color.check(
            [0, 0, 0],
        )
        color.check(
            [0, 0, 1],
        )
        color.fail([0, 0])
        color.fail([0, 0, 2])

        self.assertRaises(ValueError, color.fail, [0, 0, 1])

        # Now use ``color`` in other contracts.
        @contract
        def fill_area(inside, border):
            """Fill the area inside the current figure.

            :type border: color
            :type inside: color
            """
            pass

        @contract
        def fill_gradient(colors):
            """Use a gradient to fill the area.

            :type colors: list(color)
            """
            pass

    def test_as_decorator(self):
        @new_contract
        def even(x):
            return x % 2 == 0

        from Aspidites._vendor.contracts import parse

        p = parse("even")
        p.check(
            2,
        )
        p.check(
            4,
        )
        p.fail(3)
        p.check(
            2.0,
        )

    def test_as_decorator_multiple(self):
        @new_contract
        @contract(x="int")
        def even2(x):
            return x % 2 == 0

        from Aspidites._vendor.contracts import parse

        p = parse("even2")
        p.check(
            2,
        )
        p.fail(3)
        p.fail(2.0)  # now fails

    def test_types_as_contracts(self):
        c = cname()
        new_contract(c, str)
        check_contracts_ok(c, "")
        check_contracts_fail(c, 1)

    def test_types_as_contracts2(self):
        c = cname()
        new_contract(c, int)
        check_contracts_ok(c, 1)
        check_contracts_fail(c, "")

    def test_well_recognized(self):
        class OldStyleClass:
            def __init__(self, x, y):  # @UnusedVariable
                pass

        assert isinstance(OldStyleClass, type)

        class NewStyleClass:
            def __init__(self, x, y):  # @UnusedVariable
                pass

        assert isinstance(NewStyleClass, type)

    def test_as_decorator_with_args(self):
        from Aspidites._vendor.contracts import parse

        @new_contract
        def greater_than(value, thresh):
            return value > thresh

        p = parse("greater_than(5)")
        p.check(
            6,
        )
        p.fail(5)

    def test_as_decorator_with_kwargs(self):
        from Aspidites._vendor.contracts import parse

        @new_contract
        def less_than(value, thresh=1):
            return value < thresh

        p = parse("less_than(thresh=2)")
        p.check(
            1,
        )
        p.fail(2)

    def test_as_decorator_with_args_and_kwargs(self):
        from Aspidites._vendor.contracts import parse

        @new_contract
        def less_than_all(value, a, b, c=1, d=5):
            return value < min(a, b, c, d)

        p = parse("less_than_all(3, 4, c=5)")
        p.check(
            2,
        )
        p.fail(3)

    def test_capital_name1(self):
        # some problems with identifiers starting with capitals
        new_contract("SEn", "int")
        check_contracts_ok("SEn", 1)
        check_contracts_fail("SEn", 2.0)

    def test_capital_name2(self):
        new_contract("Sen", "int")
        check_contracts_ok("Sen", 1)
        check_contracts_fail("Sen", 2.0)


examples_valid = [
    "aa",
    "a_",
    "a2",
    "a_2",
    "list2",
    "dict2",
    "int2",
    "float2",
    "point2",
    "A2",
    "array2",
    "unit_length",
    "SE2",
    "SE3",
    "S1",
    "S2",
    "axis_angle",
]

for k in ParsingTmp.keywords:
    examples_valid.append("%s2" % k)
    examples_valid.append("%s_" % k)


def check_valid_identifier(e):
    check_valid_identifier.__dict__["description"] = "check_valid_identifier(%r)" % e

    identifier_expression.parseString(e, parseAll=True)  # @UndefinedVariable

    new_contract(e, "*")

    check(e, 42)


# def test_valid_identifiers():
#
#     for e in examples_valid:
#         yield check_valid_identifier, e



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from ..interface import Contract, ContractNotRespected
from Aspidites._vendor.contracts.syntax import W, S, Keyword, add_contract, add_keyword
from Aspidites._vendor.pyparsing import alphanums, Word

__all__ = ["IsInstance", "isinstance_contract"]


class IsInstance(Contract):

    """Checks that one of the superclasses have the specified name."""

    def __init__(self, name, where=None):
        Contract.__init__(self, where)
        self.name = name

    def check_contract(self, context, value, silent):
        class_name, bases_names = get_all_super_names(value)

        options = bases_names + [class_name]
        if not self.name in options:

            msg = "Failed check isinstance(%s) for type %r and superclasses %r." % (
                self.name,
                class_name,
                bases_names,
            )

            # Check it is just a case problem
            lc_name = self.name.lower()
            lc_options = [x.lower() for x in options]
            if lc_name in lc_options:
                msg += "\n***Note: this is just a lower/upper case error!***"

            raise ContractNotRespected(self, msg, value, context)

    def __str__(self):
        return "isinstance(%s)" % self.name

    def __repr__(self):
        return "IsInstance(%r)" % self.name

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        name = tokens["name"]
        return IsInstance(name, where)


def get_all_super_names(value):
    """Returns name of class, list of names of supers"""
    if hasattr(value, "__class__"):
        # old style class
        klass = value.__class__
        class_name = klass.__name__
        bases = get_oldstyle_bases(klass)
        bases_names = [x.__name__ for x in bases]
    else:
        # new style
        t = type(value)
        class_name = t.__name__
        bases_names = [b.__name__ for b in t.mro()]
    return class_name, bases_names


def get_oldstyle_bases(klass):
    todo = [klass]
    res = []
    while todo:
        x = todo.pop(0)
        res.append(x)
        for b in x.__bases__:
            if not b in res:
                todo.append(b)
    return res


Identifier = Word(alphanums + "_")

isinstance_contract = Keyword("isinstance") - S("(") - Identifier("name") - S(")")

add_contract(isinstance_contract.setParseAction(IsInstance.parse_action))
add_keyword("isinstance")



from Aspidites._vendor.contracts import new_contract
from Aspidites._vendor.contracts.test_registrar import fail, good, syntax_fail


@new_contract
def ext0_positive(value):
    return value > 0


@new_contract
def ext1_lessthan(value, threshold):
    return value < threshold


good("ext0_positive", 1)
fail("ext0_positive", -1)


good("ext1_lessthan(0)", -1)
fail("ext1_lessthan(0)", +1)
# named
good("ext1_lessthan(threshold=0)", -1)
fail("ext1_lessthan(threshold=0)", +1)


# needs to fail parsing because we didn't provide argument
syntax_fail("ext1_lessthan")
# needs to fail parsing because the argument name is wrong
syntax_fail("ext1_lessthan(th=0)")
# too many arguments
syntax_fail("ext1_lessthan(0,1)")



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from Aspidites._vendor.pyparsing import (
    oneOf,
    Keyword,
    replaceWith,
    Literal,
    Suppress,
    Group,
    Regex,
)

available_bool_pragmas = [
    "cython.binding",
    "cython.boundscheck",
    "cython.wraparound",
    "cython.initializedcheck",
    "cython.nonecheck",
    "cython.overflowcheck",
    "cython.overflowcheck.fold",
    "cython.embedsignature",
    "cython.cdivision",
    "cython.cdivision_warnings",
    "cython.always_allow_keywords",
    "cython.c_api_binop_methods",
    "cython.profile",
    "cython.linetrace",
    "cython.infer_types",
    "cython.type_version_tag",
    "cython.unraisable_tracebacks",
    "cython.iterable_coroutine",
    "cython.emit_code_comments",
    "cython.cpp_locals",
    "cython.optimize.use_switch",
    "cython.optimize.unpack_method_calls",
    "cython.warn.undeclared",
    "cython.warn.unreachable",
    "cython.warn.maybe_uninitialized",
    "cython.warn.unused",
    "cython.warn.unused_arg",
    "cython.warn.unused_result",
    "cython.warn.multiple_declarators",
]
available_pragmas = ["cfunc", "ccall", "nogil", "no_gc", "inline", "curried"]
indent = "    "
nl_indent = "\n" + indent
sep = ", "
pragma = Literal("#").setParseAction(replaceWith("@")).setName("pragma")
private_def_decl = (
    Literal("(").setParseAction(replaceWith("def ")).setName("function declarator")
)
args_end = (
    Group(Literal(")") + Literal(")"))
    .setParseAction(replaceWith(") -> "))
    .setName("return annotation")
)
underscore = Literal("_").setName("lambda variable")
lparen, rparen, lbrack, rbrack, lbrace, rbrace, colon, comma = map(Suppress, "()[]{}:,")
(
    lit_lparen,
    lit_rparen,
    lit_lbrack,
    lit_rbrack,
    lit_lbrace,
    lit_rbrace,
    lit_colon,
    lit_comma,
) = map(str, "()[]{}:,")
bool_literal = (
    oneOf("True False", asKeyword=True)
    .setParseAction(lambda s, l, t: t[0] == "True")
    .setName("boolean literal")
)
nullit = oneOf("/0 ø Ø").setParseAction(replaceWith("__undefined()")).setName("nullity")
signop = Literal("+") | Literal("-")
multop = Literal("*") | oneOf("/ //") | Literal("%") | Literal("@")
plusop = Literal("+") | Literal("-")
exponp = Literal("**")
factop = Literal("!")
bitwiseop = Regex(r"<<|>>|&|\||\^").setName("bitwise operator")
andop = Literal("&&").setParseAction(replaceWith(" and "))
orop = Literal("||").setParseAction(replaceWith(" or "))
comparisonop = andop | orop | Regex(">=|<=|>|<|==").setName("comparison operator")
assign_eq = Literal("=").setName("assignment")
noclosure = Literal("...").setName("ellipsis")  # That's no closure
index_items = Literal("[!]").setName("index trigram")
keys = Literal("[@]").setName("keys trigram")
persist = (
    Literal("[.]")
    .setParseAction(replaceWith(".persistent()"))
    .setName("persist trigram")
)
count_vals = Literal("[#]").setName("count trigram")
set_add = Literal("[$]").setName("set trigram")
discard = Literal("[%]").setName("discard trigram")
copy = Literal("[&]").setName("copy trigram")
append = Literal("[^]").setName("copy trigram")
update = Literal("[+]").setName("update trigram")
remove = Literal("[*]").setName("remove trigram")
match_none = (
    Literal("(!)").setParseAction(replaceWith("__match")).setName("match trigram")
)
context_stmt = (
    Literal("<!>").setParseAction(replaceWith("with ")).setName("context trigram")
)
return_none = (
    Literal("<*>").setParseAction(replaceWith("return ")).setName("return trigram")
)
yield_none = (
    Literal("<^>").setParseAction(replaceWith("yield ")).setName("yield trigram")
)
for_none = Literal("<@>").setParseAction(replaceWith("for ")).setName("loop trigram")
pass_stmt = Literal("<#>").setParseAction(replaceWith("pass")).setName("pass trigram")
cont_stmt = (
    Literal("<$>").setParseAction(replaceWith("continue")).setName("continue trigram")
)
break_stmt = (
    Literal("<%>").setParseAction(replaceWith("break")).setName("break trigram")
)
if_cond = (
    Literal("<?>").setParseAction(replaceWith("if ")).setName("conditional trigram")
)
respects = (
    Keyword("->").setParseAction(lambda s, l, t: ":").setName("contract assignment")
)
imposes = (
    Keyword("<-")
    .setParseAction(lambda s, l, t: "__new_contract")
    .setName("contract imposition")
)
struct_main = Keyword("main:").setParseAction(replaceWith('if __name__ == "__main__":'))



import unittest

from Aspidites._vendor.contracts import (
    check,
    ContractNotRespected,
    Contract,
    parse,
    check_multiple,
    ContractSyntaxError,
    fail,
)


class TestIdioms(unittest.TestCase):
    def test_check_1(self):
        res = check("tuple(int,int)", (2, 2))

        assert isinstance(res, dict)

    def test_check_1a(self):
        self.assertRaises(ValueError, check, 1, 2)

    def test_parse_1(self):
        contract = parse(">0")
        assert isinstance(contract, Contract)
        contract.check(2)
        self.assertRaises(ContractNotRespected, contract.check, "ciao")

    def test_parse_2(self):
        self.assertRaises(ContractSyntaxError, parse, ">>")

    def test_check_2(self):
        self.assertRaises(ContractNotRespected, check, "tuple(int,int)", (None, 2))

    def test_check_3(self):
        self.assertRaises(ContractSyntaxError, check, "tuple(>>int,int)", (None, 2))

    def test_check_4(self):
        score = (2, None)
        msg = "Player score must be a tuple of 2 int."
        try:
            check("tuple(int,int)", score, msg)
        except ContractNotRespected as e:
            s = str(e)
            assert msg in s
        else:
            assert False

    def test_repr_1(self):
        contract = parse(" list[N](int), N > 0")

        ("%s" % contract)  # => 'list[N](int),N>0'
        ("%r" % contract)  # => And([List(BindVariable('N',int),...

    def test_binding(self):
        context = check("list[N](str), N>0", ["a", "b", "c"])

        self.assertTrue("N" in context)
        self.assertTrue(context["N"] == 3)

    def test_check_multiple_1(self):

        data = [[1, 2, 3], [4, 5, 6]]
        row_labels = ["first season", "second season"]
        col_labels = ["Team1", "Team2", "Team3"]

        spec = [
            ("list[C](str),C>0", col_labels),
            ("list[R](str),R>0", row_labels),
            ("list[R](list[C])", data),
        ]
        check_multiple(spec)

        # now with description
        check_multiple(
            spec,
            "I expect col_labels, row_labels, data to " "have coherent dimensions.",
        )

        data = [[1, 2, 3], [1, 2]]
        spec = [
            ("list[C](str),C>0", col_labels),
            ("list[R](str),R>0", row_labels),
            ("list[R](list[C])", data),
        ]

        self.assertRaises(ContractNotRespected, check_multiple, spec)
        self.assertRaises(ContractNotRespected, check_multiple, spec, "my message")

    #     def test_symbols(self):
    #         from contracts import contract_expression  # @UnusedImport

    def test_equality_contract(self):
        c1 = parse("list[C](str),C>0")
        c2 = parse("list[C](str),C>0")
        c3 = parse("list[R](str),R>0")
        self.assertEqual(c1, c2)
        self.assertNotEqual(c1, c3)

    def test_equality_rvalue(self):
        c1 = parse("N+1")
        c2 = parse("N+2")
        c3 = parse("P+1")
        self.assertEqual(c1, c1)
        self.assertEqual(c2, c2)
        self.assertEqual(c3, c3)
        self.assertNotEqual(c1, c2)
        self.assertNotEqual(c1, c3)
        self.assertNotEqual(c2, c3)

    def test_check_context(self):
        check("N", 1, N=1)
        fail("N", 1, N=2)

        self.assertRaises(ContractNotRespected, check, "N", 1, N=2)

        self.assertRaises(ValueError, fail, "N", 1, N=1)

    def test_check_context2(self):
        """Variable names must have only one letter."""
        self.assertRaises(ValueError, check, "N", 1, NN=2)
        self.assertRaises(ValueError, check, "N", 1, nn=2)



from . import good, fail
import hypothesis
import hypothesis.strategies as st
import pytest as pt
from Aspidites._vendor.pyrsistent import pvector, pmap, pset


fail("Container", 1)
good("Hashable", 1)
fail("Iterable", 1)
fail("Sized", lambda: None)
good("Sequence", "")
good("Sized", "")
good("Callable", lambda: None)


@hypothesis.given(x=st.lists(st.text() | st.binary()))
def random_sequence(x):
    good("Container", [*x])
    good("Hashable", pvector([*x]))
    # counterexample?
    good("Iterable", [*x])
    good("Iterable", pvector([*x]))
    good("Iterator", [*x].__iter__())
    fail("Iterator", [*x])
    good("Sized", [*x])
    good("Sized", pvector([*x]))
    fail("Callable", [*x])
    good("Sequence", [*x])
    good("Sequence", pvector([*x]))
    good("Sequence", set(x))
    good("Set", {*x})
    good("Set", pset([*x]))
    fail("Set", [*x])
    good("MutableSequence", [*x])
    good("MutableSequence", pvector([*x]).evolver())
    fail("MutableSequence", (*x,))
    good("MutableSet", {*x})
    good("MutableSet", pset([*x]).evolver())
    fail("MutableSet", pset([*x]))
    fail("MutableSet", frozenset([*x]))


random_sequence()


@hypothesis.given(x=st.dictionaries(st.text(), st.binary()) | st.fixed_dictionaries({}))
def random_mappings(x):
    good("Mapping", {**x})
    good("Mapping", pmap({**x}))
    fail("Mapping", [x])
    fail("Mapping", pvector([x]))
    good("MutableMapping", {**x})
    good("MutableMapping", pmap({**x}).evolver())
    good("Iterable", pmap({**x}))
    good("Iterable", {**x})
    fail("Sequence", {**x})
    fail("Sequence", pmap({**x}))
    good("Sized", pmap({**x}))
    good("Sized", {**x})
    # good('MappingView', {}.keys())
    good("Sized", {**x}.keys())
    good("Sized", pmap({**x}).keys())
    # good('ItemsView', {}.items())
    good("Sized", {**x}.items())
    good("Sized", pmap({**x}).items())
    # good('ValuesView', {}.values())
    good("Sized", {**x}.values())
    good("Sized", pmap({**x}).values())


random_mappings()



# -*- coding: utf-8 -*-

from Aspidites._vendor.pyrsistent._pmap import pmap, m, PMap

from Aspidites._vendor.pyrsistent._pvector import pvector, v, PVector

from Aspidites._vendor.pyrsistent._pset import pset, s, PSet

from Aspidites._vendor.pyrsistent._pbag import pbag, b, PBag

from Aspidites._vendor.pyrsistent._plist import plist, l, PList

from Aspidites._vendor.pyrsistent._pdeque import pdeque, dq, PDeque

from Aspidites._vendor.pyrsistent._checked_types import (
    CheckedPMap,
    CheckedPVector,
    CheckedPSet,
    InvariantException,
    CheckedKeyTypeError,
    CheckedValueTypeError,
    CheckedType,
    optional,
)

from Aspidites._vendor.pyrsistent._field_common import (
    field,
    PTypeError,
    pset_field,
    pmap_field,
    pvector_field,
)

from Aspidites._vendor.pyrsistent._precord import PRecord

from Aspidites._vendor.pyrsistent._pclass import PClass, PClassMeta

from Aspidites._vendor.pyrsistent._immutable import immutable

from Aspidites._vendor.pyrsistent._helpers import freeze, thaw, mutant

from Aspidites._vendor.pyrsistent._transformations import inc, discard, rex, ny

from Aspidites._vendor.pyrsistent._toolz import get_in

__version__ = "0.18.0"

__all__ = (
    "pmap",
    "m",
    "PMap",
    "pvector",
    "v",
    "PVector",
    "pset",
    "s",
    "PSet",
    "pbag",
    "b",
    "PBag",
    "plist",
    "l",
    "PList",
    "pdeque",
    "dq",
    "PDeque",
    "CheckedPMap",
    "CheckedPVector",
    "CheckedPSet",
    "InvariantException",
    "CheckedKeyTypeError",
    "CheckedValueTypeError",
    "CheckedType",
    "optional",
    "PRecord",
    "field",
    "pset_field",
    "pmap_field",
    "pvector_field",
    "PClass",
    "PClassMeta",
    "immutable",
    "freeze",
    "thaw",
    "mutant",
    "get_in",
    "inc",
    "discard",
    "rex",
    "ny",
)



from __future__ import annotations

import unittest
from typing import Callable

from Aspidites._vendor.apm import *


class AggregationsTest(unittest.TestCase):
    def test_agg_set(self):
        result = match(
            [
                "foo",
                3,
                "bar",
                10,
                -4,
                "foo",
                "bar",
                17,
            ],
            Each(OneOf(InstanceOf(str) >> agg.Set("strings"), ...)),
        )
        self.assertTrue(result)
        self.assertEqual(result.strings, {"foo", "bar"})

    def test_agg_list(self):
        result = match(
            [
                "foo",
                3,
                "bar",
                10,
                -4,
                "foo",
                "bar",
                17,
            ],
            Each(OneOf(InstanceOf(str) >> agg.List("strings"), ...)),
        )
        self.assertTrue(result)
        self.assertEqual(result.strings, ["foo", "bar", "foo", "bar"])

    def test_agg_count(self):
        result = match(
            [
                1,
                2,
                3.0,
                4,
                5.0,
            ],
            Each(
                OneOf(
                    InstanceOf(int) >> agg.Count("integers"),
                    InstanceOf(float) >> agg.Count("floats"),
                )
            ),
        )
        self.assertTrue(result)
        self.assertEqual(3, result.integers)
        self.assertEqual(2, result.floats)

    def test_agg_histogram(self):
        result = match(
            ["a", "b", "c", "b", "d", "c", "a", "a", "c", "a"],
            Each(_ >> agg.Histogram("histo")),
        )
        self.assertTrue(result)
        self.assertEqual(
            {
                "a": 4,
                "b": 2,
                "c": 3,
                "d": 1,
            },
            result.histo,
        )

    def test_agg_sum(self):
        lower = 1
        upper = 10
        result = match(range(lower, upper), Each(_ >> agg.Sum("sum")))
        length = upper - lower
        self.assertTrue(result)
        self.assertEqual(length * (length + 1) // 2, result.sum)

    def test_agg_float_sum(self):
        result = match(["1", "10.3"], Each(_ >> agg.Sum("sum")))
        self.assertTrue(result)
        self.assertEqual(11.3, result.sum)

    def test_some_more_complicated_pattern(self):
        class Recursive(Pattern):
            def __init__(self, p: Callable):
                self._pattern = p

            def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
                return ctx.match(value, self._pattern())

        numbers = agg.Set()
        pattern = OneOf(
            InstanceOf(int) >> numbers, EachItem(..., Recursive(lambda: pattern))
        )

        result = match(
            {
                "a": {"b": 3},
                "c": 7,
                "d": {"e": {}},
                "f": {"g": {"h": 10, "i": 7}},
            },
            pattern,
        )
        self.assertTrue(result)
        self.assertEqual({3, 7, 10}, numbers.value)

    def test_some_more_complicated_pattern_with_nothing_in_it(self):
        class Recursive(Pattern):
            def __init__(self, p: Callable):
                self._pattern = p

            def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
                return ctx.match(value, self._pattern())

        numbers = agg.Set()
        pattern = OneOf(
            InstanceOf(float) >> numbers,
            InstanceOf(int),
            EachItem(..., Recursive(lambda: pattern)),
        )

        result = match(
            {
                "a": {"b": 3},
                "c": 7,
                "d": {"e": {}},
                "f": {"g": {"h": 10, "i": 7}},
            },
            pattern,
        )
        self.assertTrue(result)
        self.assertEqual(set(), numbers.value)



from . import agg
from .__pkginfo__ import __version__
from .case_of import case
from .core import (
    AllOf,
    Capture,
    Either,
    MatchContext,
    MatchResult,
    Not,
    OneOf,
    Pattern,
    Many,
    Strict,
    String,
    StringPattern,
    Value,
    _,
)
from .error import MatchError
from .match import match
from .overload import case_distinction, Match
from .patterns import (
    Arguments,
    At,
    Attrs,
    Between,
    Check,
    Contains,
    Each,
    EachItem,
    InstanceOf,
    IsNumber,
    IsString,
    IsTruish,
    Length,
    Maybe,
    NoneOf,
    Items,
    Regex,
    Returns,
    SubclassOf,
    Transformed,
)
from .try_match import Case, Default
from .typefoo import Parameters, KwArgs, VarArgs

Remaining = Many
Some = Many

Object = Items

__all__ = [
    "agg",
    "AllOf",
    "Capture",
    "Either",
    "Many",
    "MatchContext",
    "MatchResult",
    "Not",
    "OneOf",
    "Pattern",
    "Remaining",
    "Some",
    "Strict",
    "String",
    "StringPattern",
    "Value",
    "_",
    "case",
    "match",
    "case_distinction",
    "Match",
    "MatchError",
    "Case",
    "Default",
    "Arguments",
    "At",
    "Attrs",
    "Between",
    "Check",
    "Contains",
    "Each",
    "EachItem",
    "InstanceOf",
    "IsNumber",
    "IsString",
    "IsTruish",
    "Items",
    "Length",
    "Maybe",
    "NoneOf",
    "Object",
    "Regex",
    "Returns",
    "SubclassOf",
    "Transformed",
    "Parameters",
    "KwArgs",
    "VarArgs",
]



a: int = 3
b: str = "foo"


class MyClass:
    a: int = 4
    b: str = "bar"

    def __init__(self, a, b):
        self.a = a
        self.b = b

    def __eq__(self, other):
        return isinstance(other, MyClass) and self.a == other.a and self.b == other.b


def function(a: int, b: str) -> MyClass:
    return MyClass(a, b)


def function2(a: int, b: "str", c: MyClass) -> MyClass:
    pass


def function3(a: "int", b: "str", c: "MyClass"):
    pass


class UnannotatedClass:
    pass


def unannotated_function(a, b, c):
    pass



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8
import sys
from warnings import warn
from typing import Any, Union
from inspect import getouterframes
from math import inf, isinf, nan, isnan
from math import factorial
import numbers

try:
    import numpy as np

    Numeric = Union[int, float, complex, np.number]
except ModuleNotFoundError:
    Numeric = Union[int, float, complex]

from .api import Warn


class Undefined:
    """A monad for a failed programmatic unit; like NoneType but hashable.
    Falsy singleton acts as an absorbing element for division."""

    __slots__ = ("__weakref__", "__instance__", "func", "args", "kwargs")

    def __hash__(self):
        # noinspection PyUnresolvedReferences
        return hash(self.__weakref__)

    def __eq__(self, other: Any):
        return type(other) == Undefined

    def __add__(self, other: Any):
        return self

    def __sub__(self, other: Any):
        return self

    def __mul__(self, other: Any):
        return self

    def __truediv__(self, other: Any):
        return self

    def __floordiv__(self, other):
        return self

    def __neg__(self):
        return self

    def __invert__(self):
        return self

    def __str__(self):
        return self.__repr__()

    def __int__(self):
        return nan

    def __float__(self):
        return nan

    def __complex__(self):
        return complex(nan)

    def __oct__(self):
        return self

    def __index__(self):
        return 0

    def __len__(self):
        # Undefined has 0 elements
        return 0

    def __repr__(self):
        if hasattr(self.func, "__name__"):
            r = (
                self.__class__.__name__
                + f"({self.func.__name__}, {self.args}, {self.kwargs})"
            )
        else:
            r = self.__class__.__name__ + f"({None}, {self.args}, {self.kwargs})"
        return r

    # noinspection PyMethodMayBeStatic
    def __nonzero__(self):
        return True

    def __init__(self, func=None, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeFactorial(a: Numeric) -> Union[Numeric, Undefined]:
    if a < 0 or isnan(a) or isinf(a) or isinstance(a, (float, complex)):
        return Undefined(SafeFactorial, a)
    return factorial(a)


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeUnaryAdd(a: Numeric) -> Union[Numeric, Undefined]:
    if isnan(a) or not isinstance(a, numbers.Number):
        return Undefined(SafeUnaryAdd, a)
    return +a


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeUnarySub(a: Numeric) -> Union[Numeric, Undefined]:
    if isnan(a) or not isinstance(a, numbers.Number):
        return Undefined(SafeUnarySub, a)
    return -a


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeFloorDiv(a: Numeric, b: Numeric) -> Union[Numeric, Undefined]:
    if isinf(a) or b == 0 or (isinf(a) and isinf(b)):
        return Undefined(SafeFloorDiv, a, b)
    return a // b


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeDiv(a: Numeric, b: Numeric) -> Union[Numeric, Undefined]:
    if b == 0 or (isinf(a) and isinf(b)):
        return Undefined(SafeDiv, a, b)
    return a / b


# noinspection PyPep8Naming, PyProtectedMember,PyUnresolvedReferences
def SafeMod(a: Numeric, b: Numeric) -> Union[Numeric, Undefined]:
    if isinf(a) or b == 0:
        return Undefined(SafeMod, a, b)
    return a % b


# noinspection PyPep8Naming, PyProtectedMember,PyUnresolvedReferences
def SafeExp(a: Numeric, b: Numeric) -> Union[Numeric, Undefined]:
    if (
        (a == 0 and b == 0) or (isinf(a) and b == 0) or (isinf(b) and a == 0)
    ):  # pragma: no cover
        return Undefined(SafeExp, a, b)
    try:
        return a ** b
    except OverflowError:
        return inf  # just a really big number on most systems



from Aspidites._vendor.RestrictedPython import compile_restricted_exec
from Aspidites.tests._vendor.RestrictedPython.helper import restricted_exec


def test_RestrictingNodeTransformer__visit_AugAssign__1(mocker):
    """It allows augmented assign for variables."""
    _inplacevar_ = mocker.stub()
    _inplacevar_.side_effect = lambda op, val, expr: val + expr

    glb = {"_inplacevar_": _inplacevar_, "a": 1, "x": 1, "z": 0}

    restricted_exec("a += x + z", glb)
    assert glb["a"] == 2
    _inplacevar_.assert_called_once_with("+=", 1, 1)
    _inplacevar_.reset_mock()


def test_RestrictingNodeTransformer__visit_AugAssign__2():
    """It forbids augmented assign of attributes."""
    result = compile_restricted_exec("a.a += 1")
    assert result.errors == (
        "Line 1: Augmented assignment of attributes is not allowed.",
    )


def test_RestrictingNodeTransformer__visit_AugAssign__3():
    """It forbids augmented assign of subscripts."""
    result = compile_restricted_exec("a[a] += 1")
    assert result.errors == (
        "Line 1: Augmented assignment of object items and slices is not " "allowed.",
    )


def test_RestrictingNodeTransformer__visit_AugAssign__4():
    """It forbids augmented assign of slices."""
    result = compile_restricted_exec("a[x:y] += 1")
    assert result.errors == (
        "Line 1: Augmented assignment of object items and slices is not " "allowed.",
    )


def test_RestrictingNodeTransformer__visit_AugAssign__5():
    """It forbids augmented assign of slices with steps."""
    result = compile_restricted_exec("a[x:y:z] += 1")
    assert result.errors == (
        "Line 1: Augmented assignment of object items and slices is not " "allowed.",
    )



from __future__ import annotations

import unittest
from dataclasses import dataclass
from datetime import datetime

from Aspidites._vendor.apm import *


class BasicUseCases(unittest.TestCase):
    def test_dict(self):
        self.assertTrue(match({"foo": 1, "bar": 2, "baz": 3, "quux": 4}, {"foo": 1}))
        self.assertFalse(
            match({"foo": 1, "bar": 2, "baz": 3, "quux": 4}, Strict({"foo": 1}))
        )
        positive_number = InstanceOf(int) & Check(lambda v: v >= 0)
        self.assertTrue(
            match(
                {"foo": 1, "bar": 2},
                Strict({"foo": positive_number, "bar": positive_number}),
            )
        )

    def test_tuple(self):
        self.assertTrue(match((1, 2), (..., ...)))

    def test_strict(self):
        self.assertTrue(match(1, 1.0))
        self.assertTrue(match(1.0, Strict(1.0)))
        self.assertTrue(match(1, Strict(1)))
        self.assertFalse(match(1, Strict(1.0)))

    def test_truish(self):
        self.assertTrue(match(1, IsTruish))
        self.assertTrue(match(2.9, IsTruish))
        self.assertFalse(match("", IsTruish))

    def test_length(self):
        self.assertTrue(match([], Length(exactly=0)))
        self.assertTrue(match(("abc",), Length(exactly=1)))
        self.assertTrue(match("abc", Length(exactly=3)))
        self.assertTrue(match("abc", Length(at_least=1, at_most=4)))
        self.assertFalse(match("abc", Length(at_least=4, at_most=10)))
        self.assertTrue(match("abcde", Length(at_least=4, at_most=5)))
        self.assertFalse(match("abcdef", Length(at_least=4, at_most=5)))

    def test_contains(self):
        self.assertTrue(match([1, 2, 3], Contains(2) & Contains(3)))
        self.assertTrue(match((1, 2, 3), Contains(2) & Contains(3)))
        self.assertTrue(match({1, 2, 3}, Contains(2) & Contains(3)))
        self.assertTrue(match({1: "a", 2: "b", 3: "c"}, Contains(2) & Contains(3)))
        self.assertTrue(match("quux", Contains("uu")))

    def test_regex(self):
        self.assertTrue(match("Hello, World!", Regex("[A-Z][a-z]+, [A-Z][a-z]+!")))
        self.assertFalse(
            match(
                "Hello, World! How are you today?", Regex("[A-Z][a-z]+, [A-Z][a-z]+!")
            )
        )
        self.assertTrue(
            match(
                "Hello, World! How are you today?", Regex("[A-Z][a-z]+, [A-Z][a-z]+!.*")
            )
        )

    def test_remaining(self):
        self.assertTrue(
            match([1, 2, 3, 4], [1, 2, 3, Remaining(InstanceOf(int), at_least=1)])
        )
        self.assertTrue(match([1, 2, 3], [1, 2, 3, Remaining(InstanceOf(int))]))
        self.assertTrue(
            match(
                [1, 2, 3, 4, 5], [1, 2, 3, Remaining(InstanceOf(int) & Between(1, 10))]
            )
        )
        self.assertFalse(
            match(
                [1, 2, 3, 4, 5], [1, 2, 3, Remaining(InstanceOf(int) & Between(1, 3))]
            )
        )
        self.assertFalse(
            match([1, 2, 3, 4], [1, 2, 3, Remaining(InstanceOf(int), at_least=2)])
        )

    def test_remaining_with_range(self):
        self.assertTrue(
            match(range(1, 5), [1, 2, 3, Remaining(InstanceOf(int), at_least=1)])
        )
        self.assertTrue(match(range(1, 4), [1, 2, 3, Remaining(InstanceOf(int))]))
        self.assertTrue(
            match(range(1, 5), [1, 2, 3, Remaining(InstanceOf(int) & Between(1, 10))])
        )
        self.assertFalse(
            match(range(1, 5), [1, 2, 3, Remaining(InstanceOf(int) & Between(1, 3))])
        )
        self.assertFalse(
            match(range(1, 5), [1, 2, 3, Remaining(InstanceOf(int), at_least=2)])
        )

    def test_capture(self):
        result = match(
            {
                "This": "Is A Rather Complex Beast",
                "Created-At": "Sun Jan  3 04:08:57 CET 2021",
                "User": {
                    "UserId": 102384,
                    "FirstName": "Jane",
                    "LastName": "Doe",
                },
            },
            {
                "User": Capture(
                    {
                        "FirstName": Capture(..., name="first_name"),
                        "LastName": Capture(..., name="last_name"),
                    },
                    name="user",
                )
            },
        )
        self.assertTrue(result)
        self.assertEqual("Jane", result["first_name"])
        self.assertEqual("Doe", result["last_name"])
        self.assertEqual(102384, result["user"]["UserId"])

    def test_remaining_with_capture(self):
        result = match([1, 2, 3, 4], [1, 2, Capture(Remaining(...), name="tail")])
        self.assertTrue(result)
        self.assertEqual([3, 4], result["tail"])

    def test_and(self):
        self.assertFalse(match(0, Between(0, 1) & Between(1, 2)))
        self.assertTrue(match(1, Between(0, 1) & Between(1, 2)))
        self.assertFalse(match(2, Between(0, 1) & Between(1, 2)))

    def test_or(self):
        self.assertTrue(match(0, Between(0, 1) | Between(1, 2)))
        self.assertTrue(match(1, Between(0, 1) | Between(1, 2)))
        self.assertTrue(match(2, Between(0, 1) | Between(1, 2)))

    def test_xor(self):
        self.assertTrue(match(0, Between(0, 1) ^ Between(1, 2)))
        self.assertFalse(match(1, Between(0, 1) ^ Between(1, 2)))
        self.assertTrue(match(2, Between(0, 1) ^ Between(1, 2)))

    def test_each(self):
        self.assertTrue(match([1, 2, 3], Each(InstanceOf(int))))
        self.assertFalse(match([1, 2.0, 3], Each(InstanceOf(int))))

    def test_each_item(self):
        pattern = EachItem(Regex("[a-z]+"), InstanceOf(int))
        self.assertTrue(match({"a": 1, "b": 2}, pattern))
        self.assertFalse(match({"_a": 1, "b": 2}, pattern))
        self.assertFalse(match({"a": 7.5}, pattern))

    def test_transformed(self):
        self.assertTrue(match("value", Transformed(len, 5)))

    def test_arguments(self):
        # noinspection PyUnusedLocal
        def f(x: int, y: str):
            pass

        # noinspection PyUnusedLocal
        def g(x):
            pass

        # noinspection PyUnusedLocal
        def h(x, y: int, m: str, n: str):
            pass

        self.assertTrue(match(f, Arguments(int, str)))
        self.assertFalse(match(g, Arguments(int)))
        self.assertTrue(match(h, Arguments(None, int, Remaining(str))))

    def test_keyword_arguments(self):
        # noinspection PyUnusedLocal
        def f(x: int, y: bool):
            pass

        # noinspection PyUnusedLocal
        def g(*, x: int):
            pass

        self.assertTrue(match(f, Arguments(x=int)))
        self.assertFalse(match(f, Strict(Arguments(x=int))))
        self.assertTrue(match(f, Strict(Arguments(x=int, y=bool))))
        self.assertTrue(match(g, Arguments(x=int)))

    def test_returns(self):
        def f() -> int:
            pass

        self.assertTrue(match(f, Returns(int)))
        self.assertFalse(match(f, Returns(float)))

    def test_at(self):
        record = {"foo": {"bar": {"quux": {"value": "deeply nested"}}}}

        result = match(
            record, At("foo.bar.quux", {"value": Capture(..., name="value")})
        )
        self.assertTrue(result)
        self.assertEqual("deeply nested", result["value"])

        result = match(
            record, At(["foo", "bar", "quux"], {"value": Capture(..., name="value")})
        )
        self.assertTrue(result)
        self.assertEqual("deeply nested", result["value"])

        self.assertFalse(match(record, At("foo.bar.quux.baz", InstanceOf(int))))

    def test_string(self):
        path = "https://somehost/foo=1/bar=2/"
        result = match(
            path,
            String(
                Capture(OneOf("https", "http"), name="protocol"),
                "://",
                Capture(Regex("[a-zA-Z_-]+"), name="host"),
                Regex("/+"),
                "foo=",
                Capture(Regex("[^=/]+"), name="foo"),
                "/",
                "bar=",
                Capture(Regex("[^=/]+"), name="bar"),
                Regex("/*"),
            ),
        )
        self.assertTrue(result)
        self.assertEqual("https", result["protocol"])
        self.assertEqual("somehost", result["host"])
        self.assertEqual("1", result["foo"])
        self.assertEqual("2", result["bar"])

    def test_string_argresult(self):
        path = "https://somehost/foo=1/bar=2/"

        result = match(
            path,
            String(
                Capture(OneOf("https", "http"), name="protocol"),
                "://",
                Capture(Regex("[a-zA-Z_-]+"), name="host"),
                Regex("/+"),
                "foo=",
                Capture(Regex("[^=/]+"), name="foo"),
                "/",
                "bar=",
                Capture(Regex("[^=/]+"), name="bar"),
                Regex("/*"),
            ),
        )
        self.assertTrue(result)
        self.assertEqual("https", result.protocol)
        self.assertEqual("somehost", result.host)
        self.assertEqual("1", result.foo)
        self.assertEqual("2", result.bar)

    def test_string_with_at_capture(self):
        path = "https://somehost/foo=1/bar=2/"

        result = match(
            path,
            String(
                "protocol" @ OneOf("https", "http"),
                "://",
                "host" @ Regex("[a-zA-Z_-]+"),
                Regex("/+"),
                "foo=",
                "foo" @ Regex("[^=/]+"),
                "/",
                "bar=",
                "bar" @ Regex("[^=/]+"),
                Regex("/*"),
            ),
        )
        self.assertTrue(result)
        self.assertEqual("https", result.protocol)
        self.assertEqual("somehost", result.host)
        self.assertEqual("1", result.foo)
        self.assertEqual("2", result.bar)

    def test_is_type(self):
        self.assertTrue(match(1, IsNumber))
        self.assertTrue(match(1.0, IsNumber))
        self.assertFalse(match(True, IsNumber))

    def test_object(self):
        request = {
            "api_version": "v1",
            "job": {
                "run_at": "2020-08-27 14:09:30",
                "command": "echo 'booya'",
            },
        }

        # noinspection PyUnresolvedReferences
        result = match(
            request,
            Items(
                api_version="v1",
                job=Items(
                    run_at=Transformed(datetime.fromisoformat, "time" @ _),
                )
                & OneOf(
                    Items(command="command" @ InstanceOf(str)),
                    Items(spawn="container" @ InstanceOf(str)),
                ),
            ),
        )
        self.assertTrue(result)
        self.assertFalse("container" in result)
        self.assertEqual(result["time"], datetime(2020, 8, 27, 14, 9, 30))
        self.assertEqual(result["command"], "echo 'booya'")

    def test_attrs(self):
        @dataclass
        class Thing:
            foo: str
            bar: int

        self.assertTrue(match(Thing("xyz", 4711), Attrs(foo="xyz")))
        self.assertFalse(match(Thing("xyz", 4711), Strict(Attrs(foo="xyz"))))
        self.assertTrue(
            match(Thing("xyz", 4711), Attrs(foo=InstanceOf(str), bar=InstanceOf(int)))
        )
        self.assertTrue(
            match(
                Thing("xyz", 4711),
                Strict(Attrs(foo=InstanceOf(str), bar=InstanceOf(int))),
            )
        )

    def test_transformed_exception(self):
        request = {
            "api_version": "v1",
            "job": {
                "run_at": "junk",
                "command": "echo 'booya'",
            },
        }

        # noinspection PyUnresolvedReferences
        self.assertFalse(
            match(
                request,
                Items(
                    job=Items(
                        run_at=Transformed(datetime.fromisoformat, "time" @ _),
                    )
                ),
            )
        )

    def test_splat_result(self):
        def f(**kwargs):
            dct = {}
            for k, v in kwargs.items():
                dct[k] = v
            return dct

        result = match(
            {"a": 1, "b": 2, "c": 3}, Items(a=_ >> "x", b=_ >> "y", c=_ >> "z")
        )
        self.assertTrue(result)
        r = f(**result)
        self.assertEqual({"x": 1, "y": 2, "z": 3}, r)

    def test_maybe(self):
        pattern = Items(a=...) & Maybe(Items(b=_ >> "capture"))

        result = match({"a": 3}, pattern)
        self.assertTrue(result)
        self.assertFalse("capture" in result)
        result = match({"a": 3, "b": 5}, pattern)
        self.assertTrue(result)
        self.assertTrue("capture" in result)
        self.assertEqual(5, result["capture"])

    def test_dataclasses(self):
        @dataclass
        class User:
            first_name: str
            last_name: str

        @dataclass
        class LoggedInUser(User):
            display_name: str

        result = match(User("Jane", "Doe"), User("Jane", _ >> "last_name"))
        self.assertTrue(result)
        self.assertEqual("Doe", result["last_name"])

        result = match(
            LoggedInUser("John", "Doe", "johndoe"), User(_, _ >> "last_name")
        )
        self.assertTrue(result)
        self.assertEqual("Doe", result["last_name"])

    def test_bind(self):
        result = {}
        if match({"a": 1, "b": 2}, {"a": _ >> "foo"}).bind(result):
            self.assertEqual(1, result["foo"])
        self.assertEqual({"foo": 1}, result)

    def test_extract(self):
        # noinspection PyUnusedLocal
        def f(arg: str) -> float:
            pass

        # noinspection PyUnusedLocal
        def g(a: int, b: int):
            pass

        # noinspection PyUnresolvedReferences
        result = match(
            f, Arguments("arg_types" @ Many(...)) & Returns("return_type" @ _)
        )
        self.assertTrue(result)
        self.assertEqual([str], result.arg_types)
        self.assertEqual(float, result.return_type)

        self.assertTrue(match(f, Arguments(_)))
        self.assertFalse(match(g, Arguments(_)))
        self.assertFalse(match(f, Arguments(_, _)))
        self.assertTrue(match(g, Arguments(_, _)))

    def test_strict_match(self):
        self.assertTrue(match({"a": 3}, {"a": 3.0}))
        self.assertFalse(match({"a": 3}, {"a": 3.0}, strict=True))

    def test_groups(self):
        # noinspection PyUnresolvedReferences
        result = match([1, 2, 3], ["a" @ _, "b" @ _, "c" @ _])
        self.assertTrue(result)
        self.assertEqual(3, len(result))
        self.assertEqual({"a": 1, "b": 2, "c": 3}, result.groups())

    def test_subclass_of(self):
        result = match(bool, SubclassOf(int))
        self.assertTrue(result)

    def test_is_truish(self):
        self.assertTrue(match([0], IsTruish))
        self.assertTrue(match([], Not(IsTruish)))
        self.assertFalse(match([], IsTruish))

    def test_range(self):
        self.assertTrue(match([0, 1, 2], range(0, 3)))
        self.assertFalse(match([0, 1, 2], range(0, 3), strict=True))
        self.assertTrue(match(range(0, 3), range(0, 3), strict=True))
        self.assertFalse(match(range(0, 3), range(0, 2), strict=True))
        self.assertFalse(match(range(0, 2), range(0, 3), strict=True))



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from ..interface import Contract
from ..syntax import add_contract, W, contract_expression, Literal, Group


class SeparateContext(Contract):
    def __init__(self, contract, where=None):
        assert isinstance(contract, Contract)
        Contract.__init__(self, where)
        self.contract = contract

    def check_contract(self, context, value, silent):
        copy = context.copy()
        self.contract._check_contract(copy, value, silent)

    def __str__(self):
        return "$(%s)" % self.contract

    def __repr__(self):
        return "SeparateContext(%r)" % self.contract

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        return SeparateContext(tokens[0]["child"], where=where)


sepcon = Group(
    Literal("$") + Literal("(") - contract_expression("child") - Literal(")")
)
sepcon.setParseAction(SeparateContext.parse_action)
sepcon.setName("Context separation construct")
add_contract(sepcon)



from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval


def test_BitAnd():
    assert restricted_eval("5 & 3") == 1


def test_BitOr():
    assert restricted_eval("5 | 3") == 7


def test_BitXor():
    assert restricted_eval("5 ^ 3") == 6


def test_Invert():
    assert restricted_eval("~17") == -18


def test_LShift():
    assert restricted_eval("8 << 2") == 32


def test_RShift():
    assert restricted_eval("8 >> 1") == 4



from ._vendor.pyrsistent import PClass
from .final import final


def safer_type(*args):
    name: str
    bases: tuple
    attributes: dict
    if len(args) < 2:
        raise RuntimeError("Type introspection is not supported.")
    name, attributes = args
    bases = tuple()
    t = type(name, bases, dict(attributes))
    return final()(t)



from Aspidites.tests._vendor.RestrictedPython.helper import restricted_exec


SIMPLE_SUBSCRIPTS = """
def simple_subscript(a):
    return a['b']
"""


def test_read_simple_subscript(mocker):
    value = None
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {"_getitem_": _getitem_}
    restricted_exec(SIMPLE_SUBSCRIPTS, glb)

    assert (value, "b") == glb["simple_subscript"](value)


VAR_SUBSCRIPT = """
def simple_subscript(a, b):
    return a[b]
"""


def test_read_subscript_with_variable(mocker):
    value = [1]
    idx = 0
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {"_getitem_": _getitem_}
    restricted_exec(VAR_SUBSCRIPT, glb)

    assert (value, 0) == glb["simple_subscript"](value, idx)


TUPLE_SUBSCRIPTS = """
def tuple_subscript(a):
    return a[1, 2]
"""


def test_tuple_subscript(mocker):
    value = None
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {"_getitem_": _getitem_}
    restricted_exec(TUPLE_SUBSCRIPTS, glb)

    assert (value, (1, 2)) == glb["tuple_subscript"](value)


SLICE_SUBSCRIPT_NO_UPPER_BOUND = """
def slice_subscript_no_upper_bound(a):
    return a[1:]
"""


def test_read_slice_subscript_no_upper_bound(mocker):
    value = None
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {"_getitem_": _getitem_}
    restricted_exec(SLICE_SUBSCRIPT_NO_UPPER_BOUND, glb)

    assert (value, slice(1, None, None)) == glb["slice_subscript_no_upper_bound"](
        value
    )  # NOQA: E501


SLICE_SUBSCRIPT_NO_LOWER_BOUND = """
def slice_subscript_no_lower_bound(a):
    return a[:1]
"""


def test_read_slice_subscript_no_lower_bound(mocker):
    value = None
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {"_getitem_": _getitem_}
    restricted_exec(SLICE_SUBSCRIPT_NO_LOWER_BOUND, glb)

    assert (value, slice(None, 1, None)) == glb["slice_subscript_no_lower_bound"](
        value
    )  # NOQA: E501


SLICE_SUBSCRIPT_NO_STEP = """
def slice_subscript_no_step(a):
    return a[1:2]
"""


def test_read_slice_subscript_no_step(mocker):
    value = None
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {"_getitem_": _getitem_}
    restricted_exec(SLICE_SUBSCRIPT_NO_STEP, glb)

    assert (value, slice(1, 2, None)) == glb["slice_subscript_no_step"](value)


SLICE_SUBSCRIPT_WITH_STEP = """
def slice_subscript_with_step(a):
    return a[1:2:3]
"""


def test_read_slice_subscript_with_step(mocker):
    value = None
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {"_getitem_": _getitem_}
    restricted_exec(SLICE_SUBSCRIPT_WITH_STEP, glb)

    assert (value, slice(1, 2, 3)) == glb["slice_subscript_with_step"](value)


EXTENDED_SLICE_SUBSCRIPT = """

def extended_slice_subscript(a):
    return a[0, :1, 1:, 1:2, 1:2:3]
"""


def test_read_extended_slice_subscript(mocker):
    value = None
    _getitem_ = mocker.stub()
    _getitem_.side_effect = lambda ob, index: (ob, index)
    glb = {"_getitem_": _getitem_}
    restricted_exec(EXTENDED_SLICE_SUBSCRIPT, glb)
    ret = glb["extended_slice_subscript"](value)
    ref = (
        value,
        (
            0,
            slice(None, 1, None),
            slice(1, None, None),
            slice(1, 2, None),
            slice(1, 2, 3),
        ),
    )

    assert ref == ret


WRITE_SUBSCRIPTS = """
def assign_subscript(a):
    a['b'] = 1
"""


def test_write_subscripts(mocker):
    value = {"b": None}
    _write_ = mocker.stub()
    _write_.side_effect = lambda ob: ob
    glb = {"_write_": _write_}
    restricted_exec(WRITE_SUBSCRIPTS, glb)

    glb["assign_subscript"](value)
    assert value["b"] == 1


DEL_SUBSCRIPT = """
def del_subscript(a):
    del a['b']
"""


def test_del_subscripts(mocker):
    value = {"b": None}
    _write_ = mocker.stub()
    _write_.side_effect = lambda ob: ob
    glb = {"_write_": _write_}
    restricted_exec(DEL_SUBSCRIPT, glb)
    glb["del_subscript"](value)

    assert value == {}



from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval


def test_UAdd():
    assert restricted_eval("+a", {"a": 42}) == 42


def test_USub():
    assert restricted_eval("-a", {"a": 2411}) == -2411



"""Tests for freeze and thaw."""

from Aspidites._vendor.pyrsistent import v, m, s, freeze, thaw, PRecord, field, mutant


## Freeze (standard)


def test_freeze_basic():
    assert freeze(1) == 1
    assert freeze("foo") == "foo"


def test_freeze_list():
    assert freeze([1, 2]) == v(1, 2)


def test_freeze_dict():
    result = freeze({"a": "b"})
    assert result == m(a="b")
    assert type(freeze({"a": "b"})) is type(m())


def test_freeze_set():
    result = freeze(set([1, 2, 3]))
    assert result == s(1, 2, 3)
    assert type(result) is type(s())


def test_freeze_recurse_in_dictionary_values():
    result = freeze({"a": [1]})
    assert result == m(a=v(1))
    assert type(result["a"]) is type(v())


def test_freeze_recurse_in_pmap_values():
    input = {"a": m(b={"c": 1})}
    result = freeze(input)
    # PMap and PVector are == to their mutable equivalents
    assert result == input
    assert type(result) is type(m())
    assert type(result["a"]["b"]) is type(m())


def test_freeze_recurse_in_lists():
    result = freeze(["a", {"b": 3}])
    assert result == v("a", m(b=3))
    assert type(result[1]) is type(m())


def test_freeze_recurse_in_pvectors():
    input = [1, v(2, [3])]
    result = freeze(input)
    # PMap and PVector are == to their mutable equivalents
    assert result == input
    assert type(result) is type(v())
    assert type(result[1][1]) is type(v())


def test_freeze_recurse_in_tuples():
    """Values in tuples are recursively frozen."""
    result = freeze(("a", {}))
    assert result == ("a", m())
    assert type(result[1]) is type(m())


## Freeze (weak)


def test_freeze_nonstrict_no_recurse_in_pmap_values():
    input = {"a": m(b={"c": 1})}
    result = freeze(input, strict=False)
    # PMap and PVector are == to their mutable equivalents
    assert result == input
    assert type(result) is type(m())
    assert type(result["a"]["b"]) is dict


def test_freeze_nonstrict_no_recurse_in_pvectors():
    input = [1, v(2, [3])]
    result = freeze(input, strict=False)
    # PMap and PVector are == to their mutable equivalents
    assert result == input
    assert type(result) is type(v())
    assert type(result[1][1]) is list


## Thaw


def test_thaw_basic():
    assert thaw(1) == 1
    assert thaw("foo") == "foo"


def test_thaw_list():
    result = thaw(v(1, 2))
    assert result == [1, 2]
    assert type(result) is list


def test_thaw_dict():
    result = thaw(m(a="b"))
    assert result == {"a": "b"}
    assert type(result) is dict


def test_thaw_set():
    result = thaw(s(1, 2))
    assert result == set([1, 2])
    assert type(result) is set


def test_thaw_recurse_in_mapping_values():
    result = thaw(m(a=v(1)))
    assert result == {"a": [1]}
    assert type(result["a"]) is list


def test_thaw_recurse_in_dict_values():
    result = thaw({"a": v(1, m(b=2))})
    assert result == {"a": [1, {"b": 2}]}
    assert type(result["a"]) is list
    assert type(result["a"][1]) is dict


def test_thaw_recurse_in_vectors():
    result = thaw(v("a", m(b=3)))
    assert result == ["a", {"b": 3}]
    assert type(result[1]) is dict


def test_thaw_recurse_in_lists():
    result = thaw(v(["a", m(b=1), v(2)]))
    assert result == [["a", {"b": 1}, [2]]]
    assert type(result[0]) is list
    assert type(result[0][1]) is dict


def test_thaw_recurse_in_tuples():
    result = thaw(("a", m()))
    assert result == ("a", {})
    assert type(result[1]) is dict


def test_thaw_can_handle_subclasses_of_persistent_base_types():
    class R(PRecord):
        x = field()

    result = thaw(R(x=1))
    assert result == {"x": 1}
    assert type(result) is dict


## Thaw (weak)


def test_thaw_non_strict_no_recurse_in_dict_values():
    result = thaw({"a": v(1, m(b=2))}, strict=False)
    assert result == {"a": [1, {"b": 2}]}
    assert type(result["a"]) is type(v())
    assert type(result["a"][1]) is type(m())


def test_thaw_non_strict_no_recurse_in_lists():
    result = thaw(v(["a", m(b=1), v(2)]), strict=False)
    assert result == [["a", {"b": 1}, [2]]]
    assert type(result[0][1]) is type(m())


def test_mutant_decorator():
    @mutant
    def fn(a_list, a_dict):
        assert a_list == v(1, 2, 3)
        assert isinstance(a_dict, type(m()))
        assert a_dict == {"a": 5}

        return [1, 2, 3], {"a": 3}

    pv, pm = fn([1, 2, 3], a_dict={"a": 5})

    assert pv == v(1, 2, 3)
    assert pm == m(a=3)
    assert isinstance(pm, type(m()))



from . import syntax_fail, good, fail


### Strings
good("str", "ciao")
good("string", "ciao", exact=False)
fail("str", list("ciao"))  # sequences of chars are not str
# Can specify the length
good("str[N]", "")
good("str[1]", "a")
good("str[2]", "ab")
good("str[>0]", "ab")
fail("str[>0]", "")
good("str[N],N>3", "ciao")
fail("str[N],N>3", "cia")
# Cannot specify anything on the content
syntax_fail("str(*)")



from Aspidites._vendor.contracts import contract, new_contract, ContractNotRespected
import unittest


class ClassContractsTests(unittest.TestCase):
    def test_class_contract1(self):
        class Game(object):
            def __init__(self, legal):
                self.legal = legal

            @new_contract
            def legal_move1(self, move):
                return move in self.legal

            @contract(move="legal_move1")
            def take_turn(self, move):
                pass

        g1 = Game([1, 2])
        g1.take_turn(1)
        g1.take_turn(2)
        self.assertRaises(ContractNotRespected, g1.take_turn, 3)

    def test_class_contract2(self):
        class Game(object):
            def __init__(self, legal):
                self.legal = legal

            @new_contract
            def legal_move2(self, move):
                return move in self.legal

            @contract(move="legal_move2")
            def take_turn(self, move):
                pass

        g1 = Game([1, 2])
        g1.take_turn(1)
        g1.take_turn(2)
        self.assertRaises(ContractNotRespected, g1.take_turn, 3)

    def test_class_contract3(self):
        class Game(object):
            def __init__(self, legal):
                self.legal = legal

            def legal_move(self, move):
                return move in self.legal

            new_contract("alegalmove", legal_move)

            @contract(move="alegalmove")
            def take_turn(self, move):
                pass

        g1 = Game([1, 2])
        g1.take_turn(1)
        g1.take_turn(2)
        self.assertRaises(ContractNotRespected, g1.take_turn, 3)

    def test_class_contract1_bad(self):
        """example of bad usage, using the contract from outside"""

        class Game(object):
            def __init__(self, legal):
                self.legal = legal

            @new_contract
            def legal_move4(self, move):
                return move in self.legal

        def go():
            @contract(move="legal_move4")
            def take_turn(move):
                pass

            take_turn(0)

        self.assertRaises(ContractNotRespected, go)



from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval
import warnings


def test_Is():
    assert restricted_eval("True is True") is True


def test_NotIs():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        assert restricted_eval("1 is not True") is True



from Aspidites._vendor.RestrictedPython import compile_restricted_eval
from Aspidites._vendor._compat import IS_PY35_OR_GREATER
from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval
import pytest


# Arithmetic Operators


def test_Add():
    assert restricted_eval("1 + 1") == 2


def test_Sub():
    assert restricted_eval("5 - 3") == 2


def test_Mult():
    assert restricted_eval("2 * 2") == 4


def test_Div():
    assert restricted_eval("10 / 2") == 5


def test_Mod():
    assert restricted_eval("10 % 3") == 1


def test_Pow():
    assert restricted_eval("2 ** 8") == 256


def test_FloorDiv():
    assert restricted_eval("7 // 2") == 3


@pytest.mark.skipif(
    not IS_PY35_OR_GREATER, reason="MatMult was introducted on Python 3.5"
)
def test_MatMult():
    result = compile_restricted_eval("(8, 3, 5) @ (2, 7, 1)")
    assert result.errors == ("Line None: MatMult statements are not allowed.",)
    assert result.code is None



from . import syntax_fail, good, fail, semantic_fail
from Aspidites._vendor.contracts.library.sets import ASet

good(
    repr(ASet.__init__(ASet(), length_contract="3", elements_contract="int")), {1, 2, 3}
)
good("set[3](int)", {1, 2, 3})
fail("set[3](int)", {1, 2})
semantic_fail("set[X](int)", 1)
syntax_fail("set")



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from Aspidites._vendor.contracts.main import new_contract
from math import isfinite

try:
    import numpy  # @UnusedImport
except ImportError:  # pragma: no cover
    __all__ = ["finite"]
    new_contract("float", "Float")
    new_contract("int", "Int")
    new_contract("number", "float|int")

    @new_contract
    def finite(x):
        return isfinite(x)


else:
    __all__ = []
    new_contract("float", "Float|np_scalar_float|(np_scalar, array(float))")
    new_contract("int", "Int|np_scalar_int|(np_scalar,array(int))")
    new_contract("uint", "np_scalar_uint|(np_scalar, array(uint))")
    new_contract("number", "float|int|uint")



from __future__ import annotations

import unittest

from Aspidites._vendor.apm import *


class Glob(unittest.TestCase):
    def test_empty(self):
        self.assertTrue(match([1, 2, 3], [1, Many(...), 2, 3]))

    def test_at_least(self):
        self.assertTrue(
            match(
                [1, 2, 3],
                [
                    1,
                    Many(..., at_least=2),
                ],
            )
        )
        self.assertFalse(
            match(
                [1, 2],
                [
                    1,
                    Many(..., at_least=2),
                ],
            )
        )

    def test_at_most(self):
        self.assertTrue(
            match(
                [1, 2, 3, 4],
                [
                    1,
                    Many(..., at_most=2),
                    4,
                ],
            )
        )
        self.assertFalse(
            match(
                [1, 2, 3],
                [
                    1,
                    Many(..., at_most=2),
                    4,
                ],
            )
        )

    def test_with_remaining(self):
        self.assertTrue(
            match(
                [1, 2, 3],
                [
                    1,
                    Many(..., at_most=2),
                    Remaining(...),
                ],
            )
        )

    def test_multiple(self):
        self.assertTrue(
            match(
                range(1, 10),
                [
                    1,
                    2,
                    Many(...),
                    6,
                    Many(...),
                    8,
                    9,
                ],
            )
        )

    def test_capture(self):
        result = match(
            range(1, 10),
            [
                1,
                2,
                Capture(Many(...), name=1),
                6,
                Capture(Many(...), name=2),
                8,
                9,
            ],
        )
        self.assertTrue(result)
        self.assertEqual(result[1], [3, 4, 5])
        self.assertEqual(result[2], [7])

    def test_exactly(self):
        result = match(
            range(1, 10), [1, Many(..., exactly=3) >> "a", Many(...) >> "b", 6, 7, 8, 9]
        )
        self.assertTrue(result)
        self.assertEqual([2, 3, 4], result.a)
        self.assertEqual([5], result.b)

    def test_exactly_empty_following(self):
        result = match(
            range(1, 10), [1, Many(..., exactly=4) >> "a", Many(...) >> "b", 6, 7, 8, 9]
        )
        self.assertTrue(result)
        self.assertEqual([2, 3, 4, 5], result.a)
        self.assertEqual([], result.b)
        self.assertEqual(
            "MatchResult(matches=True, groups={'a': [2, 3, 4, 5], 'b': []})",
            repr(result),
        )



from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval
import warnings


def test_RestrictingNodeTransformer__visit_Eq__1():
    """It allows == expressions."""
    assert restricted_eval('1 == int("1")') is True


def test_RestrictingNodeTransformer__visit_NotEq__1():
    """It allows != expressions."""
    assert restricted_eval('1 != int("1")') is False


def test_RestrictingNodeTransformer__visit_Lt__1():
    """It allows < expressions."""
    assert restricted_eval("1 < 3") is True


def test_RestrictingNodeTransformer__visit_LtE__1():
    """It allows < expressions."""
    assert restricted_eval("1 <= 3") is True


def test_RestrictingNodeTransformer__visit_Gt__1():
    """It allows > expressions."""
    assert restricted_eval("1 > 3") is False


def test_RestrictingNodeTransformer__visit_GtE__1():
    """It allows >= expressions."""
    assert restricted_eval("1 >= 3") is False


def test_RestrictingNodeTransformer__visit_Is__1():
    """It allows `is` expressions."""
    assert restricted_eval("None is None") is True


def test_RestrictingNodeTransformer__visit_IsNot__1():
    """It allows `is not` expressions."""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        assert restricted_eval("2 is not None") is True


def test_RestrictingNodeTransformer__visit_In_List():
    """It allows `in` expressions for lists."""
    assert restricted_eval("2 in [1, 2, 3]") is True


def test_RestrictingNodeTransformer__visit_NotIn_List():
    """It allows `not in` expressions for lists."""
    assert restricted_eval("2 not in [1, 2, 3]") is False


def test_RestrictingNodeTransformer__visit_In_Set():
    """It allows `in` expressions for sets."""
    assert restricted_eval("2 in {1, 1, 2, 3}") is True


def test_RestrictingNodeTransformer__visit_NotIn_Set():
    """It allows `not in` expressions for sets."""
    assert restricted_eval("2 not in {1, 2, 3}") is False


def test_RestrictingNodeTransformer__visit_In_Dict():
    """It allows `in` expressions for dicts."""
    assert restricted_eval("2 in {1: 1, 2: 2, 3: 3}") is True


def test_RestrictingNodeTransformer__visit_NotIn_Dict():
    """It allows `not in` expressions for dicts."""
    assert restricted_eval("2 not in {1: 1, 2: 2, 3: 3}") is False



from . import syntax_fail, good, fail


good("map", {})
fail("map", 1)
fail("map", [])
syntax_fail("map[]")
syntax_fail("map[]()")
syntax_fail("map()")
good("map[1]", {1: 2})
good("map[N],N<2", {1: 2})
fail("map[N],N<2", {1: 2, 3: 4})
good("map(int:int)", {1: 2})
fail("map(int:int)", {"a": 2})
good("map(*:int)", {1: 2})
good("map(*:int)", {"a": 2})

# mapionary of string -> tuple, with tuple of two elements with different type
good("map(str:tuple(type(x),type(y))),x!=y", {"a": (2, 1.1)})
fail("map(str:tuple(type(x),type(y))),x!=y", {"a": (2, 1)})



__all__ = [
    "F",  # BEGIN func.py
    "curried",  # END func.py
    "take",  # BEGIN iters.py
    "drop",
    "takelast",
    "droplast",
    "consume",
    "nth",
    "first_true",
    "iterate",
    "padnone",
    "ncycles",
    "repeatfunc",
    "grouper",
    "group_by",
    "roundrobin",
    "partition",
    "splitat",
    "splitby",
    "powerset",
    "pairwise",
    "iter_suppress",
    "flatten",
    "accumulate",  # END iters.py
    "chain",
    "combinations",
    "cycle",
    "dropwhile",
    "islice",
    "repeat",
    "starmap",
    "takewhile",
    "tee",
    "call",  # BEGIN op.py
    "apply",
    "flip",
    "curry",
    "zipwith",
    "foldl",
    "foldr",
    "unfold",  # END op.py
    "tco",
    "stackless",
    "Stream",
    "_",
    "reduce",
    "filterfalse",
    "zip_longest",
    "match",
    "Capture",
    "Strict",
    "OneOf",
    "AllOf",
    "NoneOf",
    "Not",
    "Each",
    "EachItem",
    "Some",
    "Between",
    "Length",
    "Contains",
    "Regex",
    "Check",
    "InstanceOf",
    "SubclassOf",
    "Arguments",
    "Returns",
    "Transformed",
    "At",
    "Object",
]
from .fn import *
from .apm import (
    Capture,
    Strict,
    OneOf,
    AllOf,
    NoneOf,
    Not,
    Each,
    EachItem,
    Some,
    Between,
    Length,
    Contains,
    Regex,
    Check,
    InstanceOf,
    SubclassOf,
    Arguments,
    Returns,
    Transformed,
    At,
    Object,
    match,
)



# eval_arith.py
#
# Copyright 2009, 2011 Paul McGuire
#
# Expansion on the pyparsing example simpleArith.py, to include evaluation
# of the parsed tokens.
#
# Added support for exponentiation, using right-to-left evaluation of
# operands
#
from ....._vendor.pyparsing import (
    Word,
    nums,
    alphas,
    Combine,
    oneOf,
    opAssoc,
    infixNotation,
    Literal,
)


class EvalConstant(object):
    "Class to evaluate a parsed constant or variable"
    vars_ = {}

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self):
        if self.value in EvalConstant.vars_:
            return EvalConstant.vars_[self.value]
        else:
            return float(self.value)


class EvalSignOp(object):
    "Class to evaluate expressions with a leading + or - sign"

    def __init__(self, tokens):
        self.sign, self.value = tokens[0]

    def eval(self):
        mult = {"+": 1, "-": -1}[self.sign]
        return mult * self.value.eval()


def operatorOperands(tokenlist):
    "generator to extract operators and operands in pairs"
    it = iter(tokenlist)
    while 1:
        try:
            yield (next(it), next(it))
        except StopIteration:
            break


class EvalPowerOp(object):
    "Class to evaluate multiplication and division expressions"

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self):
        res = self.value[-1].eval()
        for val in self.value[-3::-2]:
            res = val.eval() ** res
        return res


class EvalMultOp(object):
    "Class to evaluate multiplication and division expressions"

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self):
        prod = self.value[0].eval()
        for op, val in operatorOperands(self.value[1:]):
            if op == "*":
                prod *= val.eval()
            if op == "/":
                prod /= val.eval()
        return prod


class EvalAddOp(object):
    "Class to evaluate addition and subtraction expressions"

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self):
        sum = self.value[0].eval()
        for op, val in operatorOperands(self.value[1:]):
            if op == "+":
                sum += val.eval()
            if op == "-":
                sum -= val.eval()
        return sum


class EvalComparisonOp(object):
    "Class to evaluate comparison expressions"
    opMap = {
        "<": lambda a, b: a < b,
        "<=": lambda a, b: a <= b,
        ">": lambda a, b: a > b,
        ">=": lambda a, b: a >= b,
        "!=": lambda a, b: a != b,
        "=": lambda a, b: a == b,
        "LT": lambda a, b: a < b,
        "LE": lambda a, b: a <= b,
        "GT": lambda a, b: a > b,
        "GE": lambda a, b: a >= b,
        "NE": lambda a, b: a != b,
        "EQ": lambda a, b: a == b,
        "<>": lambda a, b: a != b,
    }

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self):
        val1 = self.value[0].eval()
        for op, val in operatorOperands(self.value[1:]):
            fn = EvalComparisonOp.opMap[op]
            val2 = val.eval()
            if not fn(val1, val2):
                break
            val1 = val2
        else:
            return True
        return False


# define the parser
integer = Word(nums)
real = Combine(Word(nums) + "." + Word(nums))
variable = Word(alphas, exact=1)
operand = real | integer | variable

signop = oneOf("+ -")
multop = oneOf("* /")
plusop = oneOf("+ -")
expop = Literal("**")

# use parse actions to attach EvalXXX constructors to sub-expressions
operand.setParseAction(EvalConstant)
arith_expr = infixNotation(
    operand,
    [
        (signop, 1, opAssoc.RIGHT, EvalSignOp),
        (expop, 2, opAssoc.LEFT, EvalPowerOp),
        (multop, 2, opAssoc.LEFT, EvalMultOp),
        (plusop, 2, opAssoc.LEFT, EvalAddOp),
    ],
)

comparisonop = oneOf("< <= > >= != = <> LT GT LE GE EQ NE")
comp_expr = infixNotation(
    arith_expr,
    [
        (comparisonop, 2, opAssoc.LEFT, EvalComparisonOp),
    ],
)


def main():
    # sample expressions posted on comp.lang.python, asking for advice
    # in safely evaluating them
    rules = [
        "( A - B ) = 0",
        "(A + B + C + D + E + F + G + H + I) = J",
        "(A + B + C + D + E + F + G + H) = I",
        "(A + B + C + D + E + F) = G",
        "(A + B + C + D + E) = (F + G + H + I + J)",
        "(A + B + C + D + E) = (F + G + H + I)",
        "(A + B + C + D + E) = F",
        "(A + B + C + D) = (E + F + G + H)",
        "(A + B + C) = (D + E + F)",
        "(A + B) = (C + D + E + F)",
        "(A + B) = (C + D)",
        "(A + B) = (C - D + E - F - G + H + I + J)",
        "(A + B) = C",
        "(A + B) = 0",
        "(A+B+C+D+E) = (F+G+H+I+J)",
        "(A+B+C+D) = (E+F+G+H)",
        "(A+B+C+D)=(E+F+G+H)",
        "(A+B+C)=(D+E+F)",
        "(A+B)=(C+D)",
        "(A+B)=C",
        "(A-B)=C",
        "(A/(B+C))",
        "(B/(C+D))",
        "(G + H) = I",
        "-0.99 LE ((A+B+C)-(D+E+F+G)) LE 0.99",
        "-0.99 LE (A-(B+C)) LE 0.99",
        "-1000.00 LE A LE 0.00",
        "-5000.00 LE A LE 0.00",
        "A < B",
        "A < 7000",
        "A = -(B)",
        "A = C",
        "A = 0",
        "A GT 0",
        "A GT 0.00",
        "A GT 7.00",
        "A LE B",
        "A LT -1000.00",
        "A LT -5000",
        "A LT 0",
        "A=(B+C+D)",
        "A=B",
        "I = (G + H)",
        "0.00 LE A LE 4.00",
        "4.00 LT A LE 7.00",
        "0.00 LE A LE 4.00 LE E > D",
        "2**2**(A+3)",
    ]
    vars_ = {
        "A": 0,
        "B": 1.1,
        "C": 2.2,
        "D": 3.3,
        "E": 4.4,
        "F": 5.5,
        "G": 6.6,
        "H": 7.7,
        "I": 8.8,
        "J": 9.9,
    }

    # define tests from given rules
    tests = []
    for t in rules:
        t_orig = t
        t = t.replace("=", "==")
        t = t.replace("EQ", "==")
        t = t.replace("LE", "<=")
        t = t.replace("GT", ">")
        t = t.replace("LT", "<")
        t = t.replace("GE", ">=")
        t = t.replace("LE", "<=")
        t = t.replace("NE", "!=")
        t = t.replace("<>", "!=")
        tests.append((t_orig, eval(t, vars_)))

    # copy vars_ to EvalConstant lookup dict
    EvalConstant.vars_ = vars_
    failed = 0
    for test, expected in tests:
        ret = comp_expr.parseString(test)[0]
        parsedvalue = ret.eval()
        print(test, expected, parsedvalue)
        if parsedvalue != expected:
            print("<<< FAIL")
            failed += 1
        else:
            print("")

    print("")
    if failed:
        print(failed, "tests FAILED")
        return 1
    else:
        print("all tests PASSED")
        return 0


if __name__ == "__main__":
    exit(main())



try:
    import numpy
except ImportError:
    pass
else:

    import unittest

    from Aspidites._vendor.contracts import decorate, new_contract, ContractNotRespected

    new_contract("rgb", "array[HxWx3],H>0,W>0")
    new_contract("rgba", "array[HxWx4],H>0,W>0")

    def blend_function(image1, image2, bug=False):
        """
        Blends two RGB or RGBA images together.

         :param image1: The first image to blend.
         :type image1: (rgb|rgba),array[HxWx*]
         :param image2: The second image to blend.
         :type image2: (rgb|rgba),array[HxWx*]
         :param bug: Introduce a bug to check the contracts.
         :type bug: bool

         :return: The blended image.
         :rtype: rgb,array[HxWx3]
        """
        H, W = image1.shape[0], image1.shape[1]

        if bug:
            # if we want to show a bug, return a different shape
            W += 1

        result = numpy.zeros((H, W, 3), "uint8")

        # put here the actual function
        image2

        return result

    im_float = numpy.zeros((10, 10, 3), dtype="float32")
    rgb_small = numpy.zeros((10, 10, 3), dtype="uint8")
    rgb_large = numpy.zeros((20, 20, 3), dtype="uint8")

    rgba_small = numpy.zeros((10, 10, 3), dtype="uint8")
    rgba_large = numpy.zeros((20, 20, 3), dtype="uint8")

    class ArrayTest(unittest.TestCase):
        def setUp(self):
            self.blend = decorate(blend_function)

        def test_correct_behavior(self):
            self.blend(rgb_small, rgb_small)
            self.blend(rgb_small, rgba_small)
            self.blend(rgba_small, rgb_small)
            self.blend(rgb_large, rgba_large)
            self.blend(rgba_large, rgb_large)

        def test_incorrect1(self):
            self.assertRaises(ContractNotRespected, self.blend, None, None)

        def test_incorrect2(self):
            self.assertRaises(ContractNotRespected, self.blend, None, rgb_small)

        def test_incorrect3(self):
            self.assertRaises(ContractNotRespected, self.blend, rgb_small, None)

        def test_incorrect4(self):
            self.assertRaises(ContractNotRespected, self.blend, rgb_small, rgb_large)

        def test_incorrect5(self):
            self.assertRaises(ContractNotRespected, self.blend, rgb_small, rgb_large)

        def test_incorrect6(self):
            # check that rtype checking works, introduce a bug
            self.assertRaises(
                ContractNotRespected, self.blend, rgb_small, rgb_small, bug=True
            )



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from ..interface import Contract, ContractNotRespected, describe_value
from ..syntax import (
    Combine,
    Word,
    W,
    alphas,
    alphanums,
    oneOf,
    ParseException,
    ZeroOrMore,
    S,
    rvalue,
    delimitedList,
    Optional,
)
from Aspidites._vendor.pyparsing import ParseFatalException


class Extension(Contract):
    registrar = {}

    def __init__(self, identifier, where=None, args=tuple(), kwargs=None):
        assert identifier in Extension.registrar
        self.contract = Extension.registrar[identifier]
        self.identifier = identifier
        self.args = args
        self.kwargs = kwargs or {}
        Contract.__init__(self, where)

    def __str__(self):
        inside = []
        if self.args:
            inside.extend(map(str, self.args))

        if self.kwargs:
            ks = sorted(self.kwargs)
            inside.extend(["%s=%s" % (k, self.kwargs[k]) for k in ks])

        s = self.identifier

        if inside:
            return self.identifier + "(" + ",".join(inside) + ")"
        else:
            return s

    def __repr__(self):
        if self.args or self.kwargs:
            return "Extension(%r, args=%r, kwargs=%r)" % (
                self.identifier,
                self.args,
                self.kwargs,
            )

        return "Extension(%r)" % self.identifier

    def check_contract(self, context, value, silent):
        context["args"] = tuple(a.eval(context) for a in self.args)
        context["kwargs"] = dict((k, v.eval(context)) for k, v in self.kwargs.items())

        # noinspection PyProtectedMember
        self.contract._check_contract(context, value, silent)

    @staticmethod
    def parse_action(s, loc, tokens):
        identifier = tokens[0]
        args = tuple()
        kwargs = {}

        if len(tokens) == 2:
            args, kwargs = tokens[1]
            args = tuple(args)

        if not identifier in Extension.registrar:
            raise ParseException("Unknown extension contract %r" % identifier)

        # from contracts.library.separate_context import SeparateContext

        contract_ext = Extension.registrar[identifier]

        if isinstance(contract_ext, CheckCallable):
            callable_thing = contract_ext.callable

            test_args = ("value",) + args
            # don't move this import
            from Aspidites._vendor.contracts.inspection import (
                check_callable_accepts_these_arguments,
                InvalidArgs,
            )

            try:
                check_callable_accepts_these_arguments(
                    callable_thing, test_args, kwargs
                )

            except InvalidArgs as e:
                msg = "The callable %s cannot accept these arguments " % callable_thing
                msg += "args = %s, kwargs = %s " % (test_args, kwargs)
                msg += "%s" % e
                raise ParseFatalException(msg)

        where = W(s, loc)
        return Extension(identifier, where, args, kwargs)

    # We want to be pickable so we do not save self.contract
    # which might point to a lambda
    def __getstate__(self):
        return {"identifier": self.identifier, "args": self.args, "kwargs": self.kwargs}

    def __setstate__(self, d):
        self.identifier = d["identifier"]
        self.contract = Extension.registrar[self.identifier]
        self.args = d["args"]
        self.kwargs = d["kwargs"]


class CheckCallable(Contract):
    def __init__(self, callable):
        self.callable = callable
        Contract.__init__(self, where=None)

    def check_contract(self, context, value, silent):
        allowed = (ValueError, AssertionError)
        args = context.get("args", tuple())
        kwargs = context.get("kwargs", {})
        try:
            result = self.callable(value, *args, **kwargs)
        except allowed as e:  # failed
            raise ContractNotRespected(self, str(e), value, context)

        if result in [None, True]:
            # passed
            pass
        elif result == False:
            msg = "Value does not pass criteria of %s() (module: %s)." % (
                get_callable_name(self.callable),
                get_callable_module(self.callable),
            )
            raise ContractNotRespected(self, msg, value, context)
        else:
            msg = (
                "I expect that %r returns either True, False, None; or "
                "raises a ValueError exception. Instead, I got %s."
                % (self.callable, describe_value(value))
            )
            raise ValueError(msg)

    def __repr__(self):
        """Note: this contract is not representable, but anyway it is
        only used by Extension, which serializes using the identifier."""
        return "CheckCallable(%r)" % self.callable

    def __str__(self):
        """Note: this contract is not representable, but anyway it is only
        used by Extension, which serializes using the identifier."""
        return get_callable_name(callable)


def get_callable_name(c):
    """Get a displayable name for the callable even if __name__
    is not available."""
    try:
        return c.__name__ + "()"
    except AttributeError:
        return str(c)


def get_callable_module(c):
    try:
        return c.__module__
    except AttributeError:
        return "(No __module__ attr)"


def describe_callable(c):
    return get_callable_name(c) + " module: %s" % get_callable_module(c)


class CheckCallableWithSelf(Contract):
    def __init__(self, callable):  # @ReservedAssignment
        self.callable = callable
        Contract.__init__(self, where=None)

    def check_contract(self, context, value, silent):
        args = context.get("args", tuple())
        kwargs = context.get("kwargs", {})

        if not "self" in context:
            msg = (
                "You can only call this contract in the context of "
                " a function call to a regular method."
            )
            raise ContractNotRespected(self, msg, value, context)

        args = (context["self"], value) + args
        allowed = (ValueError, AssertionError)
        try:
            result = self.callable(*args, **kwargs)
        except allowed as e:  # failed
            raise ContractNotRespected(self, str(e), value, context)

        if result in [None, True]:
            # passed
            pass
        elif result == False:
            msg = "Value does not pass criteria of %s." % describe_callable(
                self.callable
            )
            raise ContractNotRespected(self, msg, value, context)
        else:
            msg = (
                "I expect that %r returns either True, False, None; or "
                "raises a ValueError exception. Instead, I got %s."
                % (self.callable, describe_value(value))
            )
            raise ValueError(msg)

    def __repr__(self):
        """Note: this contract is not representable, but anyway it is only
        used by Extension, which serializes using the identifier."""
        return "CheckCallableWithSelf(%r)" % self.callable

    def __str__(self):
        """Note: this contract is not representable, but anyway it is only
        used by Extension, which serializes using the identifier."""
        return "function %s()" % get_callable_name(self.callable)


w = Word("_" + alphanums)
arg = rvalue.copy()

kwarg = w + ZeroOrMore(" ") + S("=") + ZeroOrMore(" ") + rvalue
kwarg.setParseAction(lambda s, loc, tokens: {tokens[0]: tokens[1]})


def build_args_kwargs(s, loc, tokens):
    return (
        tuple(t for t in tokens if not isinstance(t, dict)),
        dict((k, v) for t in tokens if isinstance(t, dict) for k, v in t.items()),
    )


arglist = delimitedList(kwarg | arg)
arglist.setParseAction(build_args_kwargs)

identifier_expression = Combine(oneOf(list(alphas)) + Word("_" + alphanums)) + Optional(
    S("(") + arglist + S(")")
)


identifier_contract = identifier_expression.copy().setParseAction(
    Extension.parse_action
)



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
import argparse
import argparse as ap
import os
import sys
import traceback
import warnings
import typing as t
from pathlib import Path

from Cython import __version__ as cy_version
from Cython.Compiler import Options
from Aspidites._vendor.pyrsistent import v

import pytest

from ._vendor.semantic_version import Version
from .compiler import Compiler, CompilerArgs
from .parser import parse_module
from . import __description__

cy_version = Version.coerce(cy_version)


def get_cy_kwargs() -> dict:
    cy_opt = v(
        "annotate",
        "annotate_coverage_xml",
        "buffer_max_dims",
        "cache_builtins",
        "cimport_from_pyx",
        "clear_to_none",
        "closure_freelist_size",
        "convert_range",
        "docstrings",
        "embed_pos_in_docstring",
        "generate_cleanup_code",
        "fast_fail",
        "warning_errors",
        "error_on_unknown_names",
        "error_on_uninitialized",
        "gcc_branch_hints",
        "lookup_module_cpdef",
        "embed",
    )
    cy_kwargs = dict(zip(cy_opt, map(lambda x: getattr(Options, x), cy_opt)))
    return cy_kwargs


def get_cython_parser(
    dummy: ap.ArgumentParser,
) -> t.Tuple[ap.ArgumentParser, dict, ap.ArgumentParser, bool]:
    cy3_fallback_mode: bool = False
    cy_kwargs = get_cy_kwargs()
    if cy_version.major == 3:  # pragma: no cover
        try:
            # noinspection PyUnresolvedReferences
            from Cython.Compiler.CmdLine import create_cython_argparser

            cy_parser = create_cython_argparser()
        except Exception as e:
            warnings.warn(
                "\n"
                + "".join(traceback.format_tb(e.__traceback__))
                + "Falling back to Cython 0.X Options API",
                ImportWarning,
            )
            cy3_fallback_mode = True
            cy_parser = dummy
    else:
        cy_parser = dummy
    return dummy, cy_kwargs, cy_parser, cy3_fallback_mode


def setup_test_env(argv):
    if len(argv) >= 2 and argv[1] == "--pytest" or argv[1] == "-pt":  # pragma: no cover
        if not os.getenv("ASPIDITES_DOCKER_BUILD"):
            argv = [
                str(Path(__file__).absolute().parent / Path("tests/test_aspidites.py"))
            ] + argv[2:]
        else:
            argv = argv[2:]
        sys.exit(pytest.main(argv))


def check_noargs(argv, __test):
    if len(argv) == 1:
        from .repl import ReadEvalParse

        rep = ReadEvalParse()
        rep.loop()


def add_pre_cy3_args(parser: ap.ArgumentParser, kwargs) -> None:  # pragma: no cover
    cy_arg_group = parser.add_argument_group("optional cython arguments")
    for k, v in kwargs.items():
        cy_arg_group.add_argument(
            f'--{k.replace("_", "-")}',
            default=v,
            action="store_true" if isinstance(v, (bool,)) else "store",
        )


def parse_from_dummy(
    argv: list, dummy: ap.ArgumentParser, __test: bool = False
) -> t.Tuple[ap.Namespace, list, dict]:
    dummy, cy_kwargs, cy_parser, cy3_fallback_mode = get_cython_parser(dummy)
    check_noargs(argv, __test)
    setup_test_env(argv)
    asp_parser = ap.ArgumentParser(
        prog="aspidites",
        description=__description__,
        parents=[cy_parser],
        add_help=not bool(cy_version.major),
    )
    asp_parser.add_argument(
        "-pt", "--pytest", help="run pytest with args", metavar="ARGS"
    )
    asp_parser.add_argument("target", help="source to compile")
    asp_parser.add_argument("outpyx", help="filename to compile to"),
    # Compatible with Cython 0.X:
    # 3.0 switched to using the argparse module
    if cy_version.major == 0 or cy3_fallback_mode:
        asp_parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="forcibly overwrite existing files",
        )
        asp_parser.add_argument(
            "-p",
            "--compile-pyc",
            action="store_true",
            help="compile to python bytecode",
        )
        asp_parser.add_argument(
            "-c", "--compile-c", action="store_true", help="compile to C and run setup"
        )
        asp_parser.add_argument(
            "--build-requires",
            default="",
            metavar="",
            help="additional requirements needed to run setup (default: %(default)s)",
        )
        asp_parser.add_argument(
            "-v",
            "--verbose",
            default=0,
            action="count",
            help="increase output verbosity (default: %(default)s) e.g. -v, -vv, -vvv ",
        )
        add_pre_cy3_args(asp_parser, cy_kwargs)
        args, other_args = asp_parser.parse_known_args()
        for k in cy_kwargs.keys():
            cy_kwargs[k] = args.__getattribute__(k)
    else:
        args, other_args = asp_parser.parse_known_args()
        cy_kwargs = cy_parser.parse_known_args()
    return args, other_args, cy_kwargs


def parse_code(target, output):
    if target == "Aspidites/tests" or target == "Aspidites\\tests":  # pragma: no cover
        raise SystemExit()
    code = parse_module(open(target, "r").read())  # pragma: no cover
    if output is None:  # pragma: no cover
        output = Path(target).parent / "compiled.pyx"
    return target, output, code


def main(argv=None) -> None:
    # TODO change pyx to pyz on windows
    argv = sys.argv if not argv else argv
    # any failure results in falling back to the `Cython.Compiler.Options` API
    args, other_args, cy_kwargs = parse_from_dummy(
        argv, ap.ArgumentParser(add_help=False)
    )
    # TODO: Cython 3.0 arguments clash with Aspidites' ArgumentParser
    args.target, args.outpyx, code = parse_code(args.target, args.outpyx)
    cy_kwargs.update(
        {  # pragma: no cover
            "code": code,
            "fname": args.outpyx or "compiled.pyx",
            "force": args.force or False,
            "bytecode": args.compile_pyc,
            "c": args.compile_c,
            "build_requires": args.build_requires,
            "verbose": args.verbose,
        }
    )
    compile_args = CompilerArgs(**cy_kwargs)
    Compiler(compile_args)  # pragma: no cover



from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval


def test_Or():
    assert restricted_eval("False or True") is True


def test_And():
    assert restricted_eval("True and True") is True


def test_Not():
    assert restricted_eval("not False") is True



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

# This tiny set of safe builtins is extended by users of the module.
# AccessControl.ZopeGuards contains a large set of wrappers for builtins.
# DocumentTemplate.DT_UTil contains a few.

from Aspidites._vendor import _compat


if _compat.IS_PY2:  # pragma: no cover
    import __builtin__ as builtins
else:  # pragma: PY3
    # Do not attempt to use this package on Python2.7 as there
    # might be backports for this package such as future.
    import builtins

safe_builtins = {}

_safe_names = [
    "None",
    "False",
    "True",
    "abs",
    "bool",
    "bytes",
    "callable",
    "chr",
    "complex",
    "divmod",
    "float",
    "hash",
    "hex",
    "id",
    "int",
    "isinstance",
    "issubclass",
    "len",
    "oct",
    "ord",
    "pow",
    "range",
    "repr",
    "round",
    "slice",
    "sorted",
    "str",
    "tuple",
    "zip",
]

_safe_exceptions = [
    "ArithmeticError",
    "AssertionError",
    "AttributeError",
    "BaseException",
    "BufferError",
    "BytesWarning",
    "DeprecationWarning",
    "EOFError",
    "EnvironmentError",
    "Exception",
    "FloatingPointError",
    "FutureWarning",
    "GeneratorExit",
    "IOError",
    "ImportError",
    "ImportWarning",
    "IndentationError",
    "IndexError",
    "KeyError",
    "KeyboardInterrupt",
    "LookupError",
    "MemoryError",
    "NameError",
    "NotImplementedError",
    "OSError",
    "OverflowError",
    "PendingDeprecationWarning",
    "ReferenceError",
    "RuntimeError",
    "RuntimeWarning",
    "StopIteration",
    "SyntaxError",
    "SyntaxWarning",
    "SystemError",
    "SystemExit",
    "TabError",
    "TypeError",
    "UnboundLocalError",
    "UnicodeDecodeError",
    "UnicodeEncodeError",
    "UnicodeError",
    "UnicodeTranslateError",
    "UnicodeWarning",
    "UserWarning",
    "ValueError",
    "Warning",
    "ZeroDivisionError",
]

if _compat.IS_PY2:  # pragma: no cover
    _safe_names.extend(
        [
            "basestring",
            "cmp",
            "long",
            "unichr",
            "unicode",
            "xrange",
        ]
    )
    _safe_exceptions.extend(
        [
            "StandardError",
        ]
    )
else:  # pragma: PY3
    _safe_names.extend(
        [
            "__build_class__",  # needed to define new classes
        ]
    )

for name in _safe_names:
    safe_builtins[name] = getattr(builtins, name)

for name in _safe_exceptions:
    safe_builtins[name] = getattr(builtins, name)


# Wrappers provided by this module:
# delattr
# setattr

# Wrappers provided by ZopeGuards:
# __import__
# apply
# dict
# enumerate
# filter
# getattr
# hasattr
# iter
# list
# map
# max
# min
# sum
# all
# any

# Builtins that are intentionally disabled
# compile   - don't let them produce new code
# dir       - a general purpose introspector, probably hard to wrap
# execfile  - no direct I/O
# file      - no direct I/O
# globals   - uncontrolled namespace access
# input     - no direct I/O
# locals    - uncontrolled namespace access
# open      - no direct I/O
# raw_input - no direct I/O
# vars      - uncontrolled namespace access

# There are several strings that describe Python.  I think there's no
# point to including these, although they are obviously safe:
# copyright, credits, exit, help, license, quit

# Not provided anywhere.  Do something about these?  Several are
# related to new-style classes, which we are too scared of to support
# <0.3 wink>.  coerce, buffer, and reload are esoteric enough that no
# one should care.

# buffer
# bytearray
# classmethod
# coerce
# eval
# intern
# memoryview
# object
# property
# reload
# staticmethod
# super
# type


def _write_wrapper():
    # Construct the write wrapper class
    def _handler(secattr, error_msg):
        # Make a class method.
        def handler(self, *args):
            try:
                f = getattr(self.ob, secattr)
            except AttributeError:
                raise TypeError(error_msg)
            f(*args)

        return handler

    class Wrapper(object):
        def __init__(self, ob):
            self.__dict__["ob"] = ob

        __setitem__ = _handler(
            "__guarded_setitem__", "object does not support item or slice assignment"
        )

        __delitem__ = _handler(
            "__guarded_delitem__", "object does not support item or slice assignment"
        )

        __setattr__ = _handler(
            "__guarded_setattr__", "attribute-less object (assign or del)"
        )

        __delattr__ = _handler(
            "__guarded_delattr__", "attribute-less object (assign or del)"
        )

    return Wrapper


def _full_write_guard():
    # Nested scope abuse!
    # safetypes and Wrapper variables are used by guard()
    safetypes = {dict, list}
    Wrapper = _write_wrapper()

    def guard(ob):
        # Don't bother wrapping simple types, or objects that claim to
        # handle their own write security.
        if type(ob) in safetypes or hasattr(ob, "_guarded_writes"):
            return ob
        # Hand the object to the Wrapper instance, then return the instance.
        return Wrapper(ob)

    return guard


full_write_guard = _full_write_guard()


def guarded_setattr(object, name, value):
    setattr(full_write_guard(object), name, value)


safe_builtins["setattr"] = guarded_setattr


def guarded_delattr(object, name):
    delattr(full_write_guard(object), name)


safe_builtins["delattr"] = guarded_delattr


def safer_getattr(object, name, default=None, getattr=getattr):
    """Getattr implementation which prevents using format on string objects.

    format() is considered harmful:
    http://lucumr.pocoo.org/2016/12/29/careful-with-str-format/

    """
    if isinstance(object, _compat.basestring) and name == "format":
        raise NotImplementedError(
            "Using format() on a %s is not safe." % object.__class__.__name__
        )
    if name.startswith("_"):
        raise AttributeError(
            '"{name}" is an invalid attribute name because it '
            'starts with "_"'.format(name=name)
        )
    return getattr(object, name, default)


safe_builtins["_getattr_"] = safer_getattr


def guarded_iter_unpack_sequence(it, spec, _getiter_):
    """Protect sequence unpacking of targets in a 'for loop'.

    The target of a for loop could be a sequence.
    For example "for a, b in it"
    => Each object from the iterator needs guarded sequence unpacking.
    """
    # The iteration itself needs to be protected as well.
    for ob in _getiter_(it):
        yield guarded_unpack_sequence(ob, spec, _getiter_)


def guarded_unpack_sequence(it, spec, _getiter_):
    """Protect nested sequence unpacking.

    Protect the unpacking of 'it' by wrapping it with '_getiter_'.
    Furthermore for each child element, defined by spec,
    guarded_unpack_sequence is called again.

    Have a look at transformer.py 'gen_unpack_spec' for a more detailed
    explanation.
    """
    # Do the guarded unpacking of the sequence.
    ret = list(_getiter_(it))

    # If the sequence is shorter then expected the interpreter will raise
    # 'ValueError: need more than X value to unpack' anyway
    # => No childs are unpacked => nothing to protect.
    if len(ret) < spec["min_len"]:
        return ret

    # For all child elements do the guarded unpacking again.
    for (idx, child_spec) in spec["childs"]:
        ret[idx] = guarded_unpack_sequence(ret[idx], child_spec, _getiter_)

    return ret


safe_globals = {"__builtins__": safe_builtins}
